import hashlib

DEBUG = False

TYPE_NONE            = 'NONE'
CLIENT_DEV           = 'CDEV'
ALL_DEV              = 'ADEV'
DEVICE_LOG           = 'DLOG'
DEVICE_STATUS        = 'DSTU'
DEVICE_CMD           = 'DCMD'
DEVICE_ERASE         = 'DERS'
DEVICE_PROGRAM       = 'DPRG'
DEVICE_RESET         = 'DRST'
DEVICE_START         = 'DSTR'
DEVICE_STOP          = 'DSTP'
DEVICE_ALLOC         = 'DALC'
DEVICE_DEBUG_START   = 'DDBS'
DEVICE_DEBUG_DATA    = 'DDBD'
DEVICE_DEBUG_REINIT  = 'DDBR'
DEVICE_DEBUG_STOP    = 'DDBE'
LOG_SUB              = 'LGSB'
LOG_UNSUB            = 'LGUS'
STATUS_SUB           = 'STSB'
STATUS_UNSUB         = 'STUS'
LOG_DOWNLOAD         = 'LGDL'
FILE_BEGIN           = 'FBGN'
FILE_DATA            = 'FDTA'
FILE_END             = 'FEND'
CMD_DONE             = 'CMDD'
CMD_ERROR            = 'CMDE'
HEARTBEAT            = 'HTBT'
CLIENT_LOGIN         = 'CLGI'
TERMINAL_LOGIN       = 'TLGI'
ACCESS_LOGIN         = 'ALGI'
ACCESS_REPORT_STATUS = 'ARPS'
ACCESS_ADD_CLIENT    = 'AADC'
ACCESS_DEL_CLIENT    = 'ADLC'
ACCESS_ADD_TERMINAL  = 'AADT'
ACCESS_UPDATE_TERMINAL = 'AUPT'
ACCESS_DEL_TERMINAL  = 'ADLT'

def is_valid_type(type):
    #frequently used commands
    if type == DEVICE_LOG:
        return True
    if type == DEVICE_STATUS:
        return True
    if type == DEVICE_DEBUG_DATA:
        return True
    if type == HEARTBEAT:
        return True
    if type == DEVICE_CMD:
        return True
    if type == CMD_DONE:
        return True
    if type == CMD_ERROR:
        return True
    if type == DEVICE_ERASE:
        return True
    if type == DEVICE_PROGRAM:
        return True
    if type == DEVICE_RESET:
        return True
    if type == DEVICE_START:
        return True
    if type == DEVICE_STOP:
        return True
    if type == DEVICE_DEBUG_START:
        return True
    if type == DEVICE_DEBUG_REINIT:
        return True
    if type == DEVICE_DEBUG_STOP:
        return True
    #less frequently used commands
    if type == CLIENT_DEV:
        return True
    if type == ALL_DEV:
        return True
    if type == LOG_SUB:
        return True
    if type == LOG_UNSUB:
        return True
    if type == STATUS_SUB:
        return True
    if type == STATUS_UNSUB:
        return True
    if type == LOG_DOWNLOAD:
        return True
    if type == FILE_BEGIN:
        return True
    if type == FILE_DATA:
        return True
    if type == FILE_END:
        return True
    if type == DEVICE_ALLOC:
        return True
    if type == CLIENT_LOGIN:
        return True
    if type == TERMINAL_LOGIN:
        return True
    if type == ACCESS_LOGIN:
        return True
    if type == ACCESS_REPORT_STATUS:
        return True
    if type == ACCESS_ADD_CLIENT:
        return True
    if type == ACCESS_DEL_CLIENT:
        return True
    if type == ACCESS_ADD_TERMINAL:
        return True
    if type == ACCESS_UPDATE_TERMINAL:
        return True
    if type == ACCESS_DEL_TERMINAL:
        return True
    return False

def construct(type, value):
    if is_valid_type(type) == False:
        return ''
    if len(value) > 99999:
        print "warning: data size larger than permited"
        return ''
    frame = '{' + type + ',' + '{0:05d}'.format(len(value)) + ',' + value + '}'
    return frame

def parse(msg):
    sync = False
    type = TYPE_NONE
    length = 0
    value = ''
    while msg != '':
        if len(msg) < 12:
            type = TYPE_NONE
            length = 0
            value = ''
            break;
        #   print(msg)
        for i in range(len(msg)):
            if msg[i] != '{':
                continue
            if (i + 13) > len(msg):
                break;
            if is_valid_type(msg[i+1: i+5]) == False:
                continue
            if msg[i + 5] != ',':
                continue
            if msg[i+6 : i+11].isdigit() == False:
                continue
            if msg[i+11] != ',':
                continue
            sync = True
            if DEBUG and i > 0:
                print("msg:{0}".format(msg))
                print("discard:{0}".format(msg[0:i]))
            msg = msg[i:]
            break
        if sync == False:
            break

        type = msg[1:5]
        length = int(msg[6:11])
        if len(msg) < length + 13:
            type = TYPE_NONE
            length = 0
            value = ''
            break
        if msg[length + 12] != '}':
            sync = False
            if DEBUG: print(msg[0:12] + " Lose sync because of FOOTER error")
            msg = msg[1:]
            continue
        value = msg[12:length+12]
        msg = msg[length+13:]
        break;
    return type, length, value, msg

def hash_of_file(filename):
    h = hashlib.sha1()
    with open(filename, 'rb', buffering=0) as f:
        for b in iter(lambda : f.read(1024), b''):
            h.update(b)
    return h.hexdigest()



#!/usr/bin/env python

import sys, os, serial, time, platform, logging
from functools import partial

# MODEM Protocol bytes
NUL = b'\x00'
SOH = b'\x01'
STX = b'\x02'
EOT = b'\x04'
ACK = b'\x06'
DLE = b'\x10'
NAK = b'\x15'
CAN = b'\x18'
CRC = b'C'

LOG_LEVEL_ERROR=0
LOG_LEVEL_WARN =1
LOG_LEVEL_INFO =2
LOG_LEVEL_DEBUG=3

class XMODEM(object):
    '''
    XMODEM Protocol handler, expects an object to read from and an object to
    write to.

    >>> def getc(size, timeout=1):
    ...     return data or None
    ...
    >>> def putc(data, timeout=1):
    ...     return size or None
    ...
    >>> modem = XMODEM(getc, putc)


    :param getc: Function to retrieve bytes from a stream
    :type getc: callable
    :param putc: Function to transmit bytes to a stream
    :type putc: callable
    :param mode: XMODEM protocol mode
    :type mode: string
    :param pad: Padding character to make the packets match the packet size
    :type pad: char

    '''

    # crctab calculated by Mark G. Mendel, Network Systems Corporation
    crctable = [
        0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
        0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
        0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
        0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
        0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
        0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
        0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
        0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
        0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
        0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
        0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
        0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
        0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
        0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
        0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
        0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
        0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
        0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
        0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
        0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
        0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
        0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
        0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
        0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
        0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
        0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
        0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
        0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
        0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
        0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
        0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
        0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
    ]

    def __init__(self, getc, putc, mode='xmodem', pad=b'\x1a'):
        self.getc = getc
        self.putc = putc
        self.mode = mode
        self.pad = pad
        self.log_level = LOG_LEVEL_WARN

    def log(self, log_level, logstr):
        if log_level <= self.log_level:
            print logstr


    def abort(self, count=2, timeout=60):
        '''
        Send an abort sequence using CAN bytes.
        '''
        for _ in range(count):
            self.putc(CAN, timeout)

    def send(self, stream, retry=16, timeout=60, quiet=False, callback=None):
        '''
        Send a stream via the XMODEM protocol.

            >>> stream = open('/etc/issue', 'rb')
            >>> print(modem.send(stream))
            True

        Returns ``True`` upon successful transmission or ``False`` in case of
        failure.

        :param stream: The stream object to send data from.
        :type stream: stream (file, etc.)
        :param retry: The maximum number of times to try to resend a failed
                      packet before failing.
        :type retry: int
        :param timeout: The number of seconds to wait for a response before
                        timing out.
        :type timeout: int
        :param quiet: If True, write transfer information to stderr.
        :type quiet: bool
        :param callback: Reference to a callback function that has the
                         following signature.  This is useful for
                         getting status updates while a xmodem
                         transfer is underway.
                         Expected callback signature:
                         def callback(total_packets, success_count, error_count)
        :type callback: callable
        '''

        # initialize protocol
        try:
            packet_size = dict(
                xmodem    = 128,
                xmodem1k  = 1024,
                ymodem    = 1024,
            )[self.mode]
        except KeyError:
            raise ValueError("Invalid mode specified: {self.mode!r}"
                             .format(self=self))

            self.log(LOG_LEVEL_DEBUG, 'Begin start sequence, packet_size={0:d}'.format(packet_size))
        error_count = 0
        crc_mode = 0
        cancel = 0
        while True:
            char = self.getc(1)
            if char:
                if char == NAK:
                    self.log(LOG_LEVEL_DEBUG, 'standard checksum requested (NAK).')
                    crc_mode = 0
                    break
                elif char == CRC:
                    self.log(LOG_LEVEL_DEBUG, '16-bit CRC requested (CRC).')
                    crc_mode = 1
                    break
                elif char == CAN:
                    if not quiet:
                        sys.stderr.write('received CAN\n')
                    if cancel:
                        self.log(LOG_LEVEL_INFO, 'Transmission canceled: received 2xCAN at start-sequence')
                        return False
                    else:
                        self.log(LOG_LEVEL_DEBUG, 'cancellation at start sequence.')
                        cancel = 1
                else:
                    self.log(LOG_LEVEL_ERROR, 'send error: expected NAK, CRC, or CAN; got {0}'.format(char))

            error_count += 1
            if error_count > retry:
                self.log(LOG_LEVEL_INFO, 'send error: error_count reached {0}, aborting.'.format(retry))
                self.abort(timeout=timeout)
                return False

        # send data
        error_count = 0
        success_count = 0
        total_packets = 0
        if self.mode == 'ymodem':
            sequence = 0
            filenames = stream
        else:
            sequence = 1
        while True:
            # build packet
            if self.mode == 'ymodem' and success_count == 0:
                # send packet sequence 0 containing:
                #  Filename Length [Modification-Date [Mode [Serial-Number]]]
                # 'stream' is actually the filename
                import os
                if len(filenames):
                    filename = filenames.pop()
                    stream = open(filename, 'rb')
                    stat = os.stat(filename)
                    data = os.path.basename(filename) + NUL + str(stat.st_size)
                    self.log(LOG_LEVEL_DEBUG, 'ymodem sending : {0} len:{1}'.format(filename, stat.st_size))
                else:
                    # empty file name packet terminates transmission
                    filename = ''
                    data = ''
                    stream = None
                    self.log(LOG_LEVEL_DEBUG, 'ymodem done, sending empty header.')
                if len(data) <= 128:
                    header_size = 128
                else:
                    header_size = 1024

                header = self._make_send_header(header_size, sequence)
                data = data.ljust(header_size, NUL)
                checksum = self._make_send_checksum(crc_mode, data)
            else:
                # happens after sending ymodem empty filename
                if not stream:
                    return True
                # normal data packet
                data = stream.read(packet_size)
                if not data:
                    # end of stream
                    self.log(LOG_LEVEL_DEBUG, 'send: at EOF')
                    break
                total_packets += 1

                header = self._make_send_header(packet_size, sequence)
                data = data.ljust(packet_size, self.pad)
                checksum = self._make_send_checksum(crc_mode, data)

            # emit packet & get ACK
            while True:
                self.log(LOG_LEVEL_DEBUG, 'send: block {0}'.format(sequence))
                self.putc(header + data + checksum)
                char = self.getc(1, timeout)
                if char == ACK:
                    success_count += 1
                    if callable(callback):
                        callback(total_packets, success_count, error_count)
                    error_count = 0
                    if self.mode == 'ymodem' and success_count == 1 and len(filename):
                        char = self.getc(1, timeout)
                        if char == DLE: # dunno why
                            char = self.getc(1, timeout)
                        if char == CRC:
                            break
                        self.log(LOG_LEVEL_ERROR, 'send error: ymodem expected CRC; got {0} for block {1}'.format(char, sequence))
                    else:
                        break

                self.log(LOG_LEVEL_ERROR, 'send error: expected ACK; got {0} for block {1}'.format(char, sequence))
                error_count += 1
                if callable(callback):
                    callback(total_packets, success_count, error_count)
                if error_count > retry:
                    # excessive amounts of retransmissions requested,
                    # abort transfer
                    self.log(LOG_LEVEL_ERROR, 'send error: NAK received {0} times, aborting.'.format(error_count))
                    self.abort(timeout=timeout)
                    return False

            # keep track of sequence
            sequence = (sequence + 1) % 0x100

        # emit EOT and get corresponding ACK
        while True:
            self.log(LOG_LEVEL_DEBUG, 'sending EOT, awaiting ACK')
            # end of transmission
            self.putc(EOT)

            # An ACK should be returned
            char = self.getc(1, timeout)
            if char == ACK:
                break
            else:
                self.log(LOG_LEVEL_ERROR, 'send error: expected ACK; got {0}'.format(char))
                error_count += 1
                if error_count > retry:
                    self.log(LOG_LEVEL_WARN, 'EOT was not ACKd, aborting transfer')
                    self.abort(timeout=timeout)
                    return False

        self.log(LOG_LEVEL_INFO, 'Transmission successful (ACK received).')
        if self.mode == 'ymodem':
            # YMODEM - recursively send next file
            # or empty filename header to end the xfer batch.
            stream.close()
            return self.send(filenames, retry, timeout, quiet, callback)
        return True

    def _make_send_header(self, packet_size, sequence):
        assert packet_size in (128, 1024), packet_size
        _bytes = []
        if packet_size == 128:
            _bytes.append(ord(SOH))
        elif packet_size == 1024:
            _bytes.append(ord(STX))
        _bytes.extend([sequence, 0xff - sequence])
        return bytearray(_bytes)

    def _make_send_checksum(self, crc_mode, data):
        _bytes = []
        if crc_mode:
            crc = self.calc_crc(data)
            _bytes.extend([crc >> 8, crc & 0xff])
        else:
            crc = self.calc_checksum(data)
            _bytes.append(crc)
        return bytearray(_bytes)

    def recv(self, stream, crc_mode=1, retry=16, timeout=60, delay=1, quiet=0):
        '''
        Receive a stream via the XMODEM protocol.

            >>> stream = open('/etc/issue', 'wb')
            >>> print(modem.recv(stream))
            2342

        Returns the number of bytes received on success or ``None`` in case of
        failure.
        '''

        # initiate protocol
        error_count = 0
        char = 0
        cancel = 0
        while True:
            # first try CRC mode, if this fails,
            # fall back to checksum mode
            if error_count >= retry:
                self.log(LOG_LEVEL_INFO, 'error_count reached {0}, aborting.'.format(retry))
                self.abort(timeout=timeout)
                return None
            elif crc_mode and error_count < (retry // 2):
                if not self.putc(CRC):
                    self.log(LOG_LEVEL_DEBUG, 'recv error: putc failed, sleeping for {0}'.format(delay))
                    time.sleep(delay)
                    error_count += 1
            else:
                crc_mode = 0
                if not self.putc(NAK):
                    self.log(LOG_LEVEL_DEBUG, 'recv error: putc failed, sleeping for {0}'.format(delay))
                    time.sleep(delay)
                    error_count += 1

            char = self.getc(1, timeout)
            if char is None:
                self.log(LOG_LEVEL_WARN, 'recv error: getc timeout in start sequence')
                error_count += 1
                continue
            elif char == SOH:
                self.log(LOG_LEVEL_DEBUG, 'recv: SOH')
                break
            elif char == STX:
                self.log(LOG_LEVEL_DEBUG, 'recv: STX')
                break
            elif char == CAN:
                if cancel:
                    self.log(LOG_LEVEL_INFO, 'Transmission canceled: received 2xCAN at start-sequence')
                    return None
                else:
                    self.log(LOG_LEVEL_DEBUG, 'cancellation at start sequence.')
                    cancel = 1
            else:
                error_count += 1

        # read data
        error_count = 0
        income_size = 0
        packet_size = 128
        sequence = 1
        cancel = 0
        while True:
            while True:
                if char == SOH:
                    if packet_size != 128:
                        self.log(LOG_LEVEL_DEBUG, 'recv: SOH, using 128b packet_size')
                        packet_size = 128
                    break
                elif char == STX:
                    if packet_size != 1024:
                        self.log(LOG_LEVEL_DEBUG, 'recv: SOH, using 1k packet_size')
                        packet_size = 1024
                    break
                elif char == EOT:
                    # We received an EOT, so send an ACK and return the
                    # received data length.
                    self.putc(ACK)
                    self.log(LOG_LEVEL_INFO, "Transmission complete, {0} bytes".format(income_size))
                    return income_size
                elif char == CAN:
                    # cancel at two consecutive cancels
                    if cancel:
                        self.log(LOG_LEVEL_INFO, 'Transmission canceled: received 2xCAN at block {0}'.format(sequence))
                        return None
                    else:
                        self.log(LOG_LEVEL_DEBUG, 'cancellation at block {0}'.format(sequence))
                        cancel = 1
                else:
                    err_msg = ('recv error: expected SOH, EOT; got {0}'.format(char))
                    if not quiet:
                        sys.stderr.write(err_msg+"\n")
                    self.log(LOG_LEVEL_WARN, err_msg)
                    error_count += 1
                    if error_count > retry:
                        self.log(LOG_LEVEL_INFO, 'error_count reached {0}, aborting.'.format(retry))
                        self.abort()
                        return None

            # read sequence
            error_count = 0
            cancel = 0
            self.log(LOG_LEVEL_DEBUG, 'recv: data block {0}'.format(sequence))
            seq1 = self.getc(1, timeout)
            if seq1 is None:
                self.log(LOG_LEVEL_WARN, 'getc failed to get first sequence byte')
                seq2 = None
            else:
                seq1 = ord(seq1)
                seq2 = self.getc(1, timeout)
                if seq2 is None:
                    self.log(LOG_LEVEL_WARN, 'getc failed to get second sequence byte')
                else:
                    # second byte is the same as first as 1's complement
                    seq2 = 0xff - ord(seq2)

            if not (seq1 == seq2 == sequence):
                # consume data anyway ... even though we will discard it,
                # it is not the sequence we expected!
                err_msg = 'expected sequence {0}, got (seq1={1}, seq2={2}), \
                          receiving next block, will NAK.'.format(sequence, seq1, seq2)
                self.log(LOG_LEVEL_ERROR, err_msg)
                self.getc(packet_size + 1 + crc_mode)
            else:
                # sequence is ok, read packet
                # packet_size + checksum
                data = self.getc(packet_size + 1 + crc_mode, timeout)
                valid, data = self._verify_recv_checksum(crc_mode, data)

                # valid data, append chunk
                if valid:
                    income_size += len(data)
                    stream.write(data)
                    self.putc(ACK)
                    sequence = (sequence + 1) % 0x100
                    # get next start-of-header byte
                    char = self.getc(1, timeout)
                    continue

            # something went wrong, request retransmission
            self.log(LOG_LEVEL_WARN, 'recv error: purge, requesting retransmission (NAK)')
            while True:
                # When the receiver wishes to <nak>, it should call a "PURGE"
                # subroutine, to wait for the line to clear. Recall the sender
                # tosses any characters in its UART buffer immediately upon
                # completing sending a block, to ensure no glitches were mis-
                # interpreted.  The most common technique is for "PURGE" to
                # call the character receive subroutine, specifying a 1-second
                # timeout, and looping back to PURGE until a timeout occurs.
                # The <nak> is then sent, ensuring the other end will see it.
                data = self.getc(1, timeout=1)
                if data is None:
                    break
                assert False, data
            self.putc(NAK)
            # get next start-of-header byte
            char = self.getc(1, timeout)
            continue

    def _verify_recv_checksum(self, crc_mode, data):
        if crc_mode:
            _checksum = bytearray(data[-2:])
            their_sum = (_checksum[0] << 8) + _checksum[1]
            data = data[:-2]

            our_sum = self.calc_crc(data)
            valid = bool(their_sum == our_sum)
            if not valid:
                self.log(LOG_LEVEL_WARN, 'recv error: checksum fail (theirs={0:04x}, ours={1:04x}), '.format(their_sum, our_sum))
        else:
            _checksum = bytearray([data[-1]])
            their_sum = _checksum[0]
            data = data[:-1]

            our_sum = self.calc_checksum(data)
            valid = their_sum == our_sum
            if not valid:
                self.log(LOG_LEVEL_WARN, 'recv error: checksum fail (theirs={0:02x}, ours={1:02x})'.forma(their_sum, our_sum))
        return valid, data

    def calc_checksum(self, data, checksum=0):
        '''
        Calculate the checksum for a given block of data, can also be used to
        update a checksum.

            >>> csum = modem.calc_checksum('hello')
            >>> csum = modem.calc_checksum('world', csum)
            >>> hex(csum)
            '0x3c'

        '''
        if platform.python_version_tuple() >= ('3', '0', '0'):
            return (sum(data) + checksum) % 256
        else:
            return (sum(map(ord, data)) + checksum) % 256

    def calc_crc(self, data, crc=0):
        '''
        Calculate the Cyclic Redundancy Check for a given block of data, can
        also be used to update a CRC.

            >>> crc = modem.calc_crc('hello')
            >>> crc = modem.calc_crc('world', crc)
            >>> hex(crc)
            '0xd5e3'

        '''
        for char in bytearray(data):
            crctbl_idx = ((crc >> 8) ^ char) & 0xff
            crc = ((crc << 8) ^ self.crctable[crctbl_idx]) & 0xffff
        return crc & 0xffff

port = 0
prcss_debug = False
modem_debug = False

def getc(size, timeout=1):
    port.timeout = timeout
    ret = port.read(size)
    if modem_debug:
        print "read:",
        print ret,
        print ", {0} bytes".format(len(ret))
    return ret

def putc(data, timeout=1):
    port.timeout = timeout
    len = port.write(data)
    if modem_debug:
        print "{0} bytes writen".format(len)
    return len

def send_file(filename):
    modem = XMODEM(getc, putc, mode="ymodem")
    return modem.send([filename])

def assert_response(patterns, timeout):
    port.timeout = 0.01
    timeout_tick = time.time() + timeout
    while True:
        line = port.readline()
        if prcss_debug and len(line):
            print line
        for pattern in patterns:
            if pattern in line:
                return True
        if time.time() > timeout_tick:
            return False

def print_usage():
    print "Usage: {0} port [-a app.bin] [-b bootloader.bin] [-d driver.bin] [--bootloader-baudrate 921600] [--application-baudrate 115200] [--noboot]\n".format(sys.argv[0])
    print "  examples: python {0} /dev/ttyUSB0 -a app.bin, to update app only".format(sys.argv[0])
    print "          : python {0} /dev/ttyUSB1 -b bootloader.bin -a app.bin, to update bootloader and app".format(sys.argv[0])
    print "          : python {0} /dev/ttyUSB0 -a app.bin -d driver.bin, to update app and driver".format(sys.argv[0])

if len(sys.argv) < 4:
    print_usage()
    sys.exit(1)

device = sys.argv[1]

updates=[]
bootloader_baudrate=921600
application_baudrate=921600
bootapp = True
hardreboot = False

i = 2
update = 0
while i < len(sys.argv):
    if sys.argv[i].startswith("0x") and (i + 1) < len(sys.argv):
        if os.path.isfile(sys.argv[i+1]) == False:
            sys.stderr.write("error: file {0} does not exist\n".format(sys.argv[i+1]))
            sys.exit(1)
        updates.append([sys.argv[i], sys.argv[i+1]])
        update += 1
        i += 1
    elif sys.argv[i] == "--bootloader-baudrate" and (i + 1) < len(sys.argv):
        try:
            bootloader_baudrate = int(sys.argv[i+1])
        except:
            sys.stderr.write("error: invalid bootload baudrate value {0}\n".format(sys.argv[i+1]))
            sys.exit(1)
        i += 1
    elif sys.argv[i] == "--application-baudrate" and (i + 1) < len(sys.argv):
        try:
            application_baudrate = int(sys.argv[i+1])
        except:
            sys.stderr.write("error: invalid bootload baudrate value {0}\n".format(sys.argv[i+1]))
            sys.exit(1)
        i += 1
    elif sys.argv[i] == "--hardreset":
        hardreboot = True
    elif sys.argv[i] == "--noappboot":
        bootapp = False
    i += 1

if update <= 0:
    sys.exit(0)

try:
    port = serial.Serial(device, bootloader_baudrate, timeout = 0.05)
    port.setRTS(False)
except:
    sys.stderr.write("error: unable to open {0}\n".format(device))
    sys.exit(1)

if hardreboot == False:
    port.write("a\r\n") #abort potential ongoing YMODEM transfer
    port.flushInput()
    port.write("help\r\n")
    if assert_response(["bootloader", "read [address] [size]"], 1) == False:
        if application_baudrate != bootloader_baudrate:
            port.baudrate = application_baudrate
            port.flushInput()
            port.write("dummycmd_for_flushing_purpose\r\n")
            time.sleep(0.1)
        port.write("reboot\r\n")
        if assert_response(["reboot"] , 1) == False:
            sys.stderr.write("error: failed to reboot the board, it did not respond to \"reboot\" command\n")
            sys.exit(1)
        if application_baudrate != bootloader_baudrate:
            port.baudrate = bootloader_baudrate
        time.sleep(0.11); port.write("   ") #0.11s
        time.sleep(0.03); port.write("   ") #0.14s
        time.sleep(0.03); port.write("   ") #0.17s
        time.sleep(0.03); port.write("   ") #0.20s
        time.sleep(0.03); port.write("   ") #0.23s
        time.sleep(0.03); port.write("   \r\n") #0.26s
        if assert_response(["ootloader", "BOOTLODER"], 1) == False:
            sys.stderr.write("error: failed to enter bootloader\n")
            sys.exit(1)
else:
    port.setRTS(True)
    time.sleep(0.1)
    port.setRTS(False)
    time.sleep(0.11); port.write("   ") #0.11s
    time.sleep(0.03); port.write("   ") #0.14s
    time.sleep(0.03); port.write("   ") #0.17s
    time.sleep(0.03); port.write("   ") #0.20s
    time.sleep(0.03); port.write("   ") #0.23s
    time.sleep(0.03); port.write("   \r\n") #0.26s
    if assert_response(["ootloader"], 1) == False:
        sys.stderr.write("error: failed to enter bootloader\n")
        sys.exit(1)
port.flushInput()

failed_num = 0
for [addr, image] in updates:
    status_str = "updating {0} with {1} @ address {2} ...".format(device, image, addr)
    print status_str
    port.write("write {0}\r\n".format(addr))
    if assert_response(["Waiting for the file to be sent"], 1) == False:
        sys.stderr.write("error: waiting for target to enter into YMODEM recived mode failed\n")
        sys.exit(1)
    result = send_file(image)
    if result == True:
        status_str = status_str + " succeed"
    else:
        status_str = status_str + " failed"
        failed_num += 1
    print status_str

if bootapp and failed_num == 0:
    port.write("boot\r\n")
    assert_response(["Booting......"], 1)
port.close()
sys.exit(failed_num)




#!/usr/bin/python

import os, sys, time, re, random, json, Queue
import curses, socket, ssl, select
import subprocess, thread, threading
import traceback
from operator import itemgetter
from os import path
import packet as pkt

MAX_MSG_LENGTH = 65536
CMD_WINDOW_HEIGHT = 2
DEV_WINDOW_WIDTH  = 36
LOG_WINDOW_HEIGHT = 30
LOG_WINDOW_WIDTH  = 80
LOG_HISTORY_LENGTH = 5000
MOUSE_SCROLL_UP = 0x80000
MOUSE_SCROLL_DOWN = 0x8000000
ENCRYPT = True
DEBUG = True

class ConnectionLost(Exception):
    pass

class Terminal:
    def __init__(self):
        self.stdscr = 0
        self.log_window = 0
        self.dev_window = 0
        self.cmd_window = 0
        self.keep_running = True
        self.connected = False
        self.device_list= {}
        self.service_socket = None
        self.cmd_excute_state = 'idle'
        self.cmd_excute_return = ''
        self.cmd_excute_event = threading.Event()
        self.output_queue = Queue.Queue(256)
        self.debug_socks = {}
        self.debug_queue = Queue.Queue(256)
        self.log_content = []
        self.log_curr_line = -1
        self.log_subscribed = []
        self.using_list = []
        self.max_log_width = 0
        self.cur_color_pair = 7
        self.alias_tuples = {}
        self.cmd_history = []
        self.last_runcmd_dev = []
        self.uuid = None

    def init(self):
        work_dir = path.join(path.expanduser('~'), '.udterminal')
        if path.exists(work_dir) == False:
            os.mkdir(work_dir)

        alias_file = path.join(path.expanduser('~'), '.udterminal', '.alias')
        if path.exists(alias_file) == True:
            try:
                file = open(alias_file, 'rb')
                self.alias_tuples = json.load(file)
                file.close()
            except:
                print 'read alias record failed'

        cmd_history_file = path.join(path.expanduser('~'), '.udterminal', '.cmd_history')
        if path.exists(cmd_history_file) == True:
            try:
                file = open(cmd_history_file, 'rb')
                self.cmd_history = json.load(file)
                file.close()
            except:
                print 'read command history failed'

        accesskey_file = path.join(path.expanduser('~'), '.udterminal', '.accesskey')
        if path.exists(accesskey_file) == True:
            try:
                file = open(accesskey_file, 'rb')
                self.uuid = json.load(file)
                file.close()
            except:
                print 'read access key failed'
        while self.uuid == None:
            try:
                uuid = raw_input('please input your access key: ')
            except:
                return 'read access key fail'
            is_valid_uuid = re.match('^[0-9a-f]{10,20}$', uuid)
            if is_valid_uuid == None:
                print 'invalid access key {0}, please enter a valid access key'.format(uuid)
                continue
            self.uuid = uuid
            try:
                accesskey_file = path.join(path.expanduser('~'), '.udterminal', '.accesskey')
                file = open(accesskey_file, 'wb')
                json.dump(self.uuid, file)
                file.close()
            except:
                print 'error: save access key to file failed'

        #initialize UI
        try:
            self.stdscr = curses.initscr()
            curses.noecho()
            curses.cbreak()
            curses.mousemask(-1)
            curses.mouseinterval(0)
            curses.start_color()
            curses.use_default_colors()
            curses.init_pair(1, curses.COLOR_RED, -1)
            curses.init_pair(2, curses.COLOR_GREEN, -1)
            curses.init_pair(3, curses.COLOR_YELLOW, -1)
            curses.init_pair(4, curses.COLOR_BLUE, -1)
            curses.init_pair(5, curses.COLOR_MAGENTA, -1)
            curses.init_pair(6, curses.COLOR_CYAN, -1)
            curses.init_pair(7, curses.COLOR_WHITE, -1)
            self.curseslock = threading.Lock();
            self.window_redraw()
            return 'success'
        except:
            if DEBUG: traceback.print_exc()
            return 'UI failed'

    def window_redraw(self):
        global LOG_WINDOW_WIDTH
        global LOG_WINDOW_HEIGHT
        self.curseslock.acquire()
        (max_y, max_x) = self.stdscr.getmaxyx()
        for y in range(max_y):
            for x in range(max_x - 1):
                self.stdscr.addch(y, x, ' ')
        self.stdscr.refresh()
        LOG_WINDOW_HEIGHT = max_y - CMD_WINDOW_HEIGHT - 3
        LOG_WINDOW_WIDTH = max_x - DEV_WINDOW_WIDTH - 3
        width = 1 + LOG_WINDOW_WIDTH + 1 + DEV_WINDOW_WIDTH + 1
        height = 1 + LOG_WINDOW_HEIGHT + 1 + CMD_WINDOW_HEIGHT + 1
        horline = '-' * (width-2)
        self.stdscr.addstr(0, 1, horline)
        self.stdscr.addstr(LOG_WINDOW_HEIGHT + 1, 1, horline)
        self.stdscr.addstr(LOG_WINDOW_HEIGHT + CMD_WINDOW_HEIGHT + 2, 1, horline)
        for i in range(1, height-1):
            self.stdscr.addch(i, 0, '|')
        for i in range(1, height - CMD_WINDOW_HEIGHT - 2):
            self.stdscr.addch(i, LOG_WINDOW_WIDTH + 1, '|')
        for i in range(1, height-1):
            self.stdscr.addch(i, LOG_WINDOW_WIDTH + DEV_WINDOW_WIDTH + 2, '|')
        self.stdscr.refresh()
        x = 1; y = 1;
        self.log_window = curses.newwin(LOG_WINDOW_HEIGHT, LOG_WINDOW_WIDTH, y, x)
        x = 1 + LOG_WINDOW_WIDTH + 1; y = 1;
        self.dev_window = curses.newwin(LOG_WINDOW_HEIGHT, DEV_WINDOW_WIDTH, y, x)
        x = 1; y = 1 + LOG_WINDOW_HEIGHT + 1;
        self.cmd_window = curses.newwin(CMD_WINDOW_HEIGHT, LOG_WINDOW_WIDTH + 1 + DEV_WINDOW_WIDTH, y, x)
        self.cmd_window.keypad(1)
        self.log_window.addstr(0, 0, 'Logs', curses.A_BOLD)
        self.log_window.move(1, 0)
        self.log_window.refresh()
        self.dev_window.addstr(0, 0, 'Devices', curses.A_BOLD)
        self.dev_window.addstr(0, DEV_WINDOW_WIDTH-4, 'USE')
        self.dev_window.move(1, 0)
        self.dev_window.refresh()
        self.cmd_window.addstr(0, 0, 'Command:', curses.A_BOLD)
        self.cmd_window.refresh()
        self.curseslock.release()

    def process_esc_sequence(self, escape):
        colors = escape[2:-1].split(';')
        self.cur_color_pair = 7
        for color in colors:
            if color == '30':
                self.cur_color_pair = 0
            elif color == '31':
                self.cur_color_pair = 1
            elif color == '32':
                self.cur_color_pair = 2
            elif color == '33':
                self.cur_color_pair = 3
            elif color == '34':
                self.cur_color_pair = 4
            elif color == '35':
                self.cur_color_pair = 5
            elif color == '36':
                self.cur_color_pair = 6
            elif color == '37':
                self.cur_color_pair = 7

    def log_parse_oneline(self, log):
        log_buffer = []

        if ':' in log:
            log_index = log.split(':')[0] + ':'
        else:
            log_index = ''
        log = log[len(log_index):]
        log = log.replace('\t', ' ' * 8)
        line_length = len(log)
        j = 0; log_len = 0; log_str = log_index;
        while j < line_length:
            c = log[j]
            if c == '\x1B' and log[j:j+2] == '\x1B[': #ESC
                k = j + 2
                find_esc_seq = False
                while k < line_length:
                    if log[k] == 'm':
                        find_esc_seq = True
                        break
                    k += 1
                if find_esc_seq:
                    self.process_esc_sequence(log[j:k+1])
                    j = k + 1
                    continue
            if ord(c) >= 0x20 and ord(c) < 0x7F:
                log_str += c
                log_len += 1
                if log_len + len(log_index) >= LOG_WINDOW_WIDTH - 1:
                    log_buffer.append([self.cur_color_pair, log_str])
                    log_len = 0; log_str = log_index
            j += 1

        if log_len > 0:
            log_buffer.append([self.cur_color_pair, log_str])

        if len(log_buffer) == 0:
            log_buffer.append([self.cur_color_pair, log_index])

        return log_buffer

    def log_display(self, logtime, log):
        #save log
        if log != '':
            self.log_content.append((logtime, log))
            self.log_content.sort(key=itemgetter(0))
            if len(self.log_content) > LOG_HISTORY_LENGTH:
                self.log_content.pop(0)
                if self.log_curr_line > 0:
                    self.log_curr_line -= 1

        if self.log_curr_line != -1 and log != '':
            return

        #clear log screen
        self.log_window.move(1,0)
        clear = (' ' * (LOG_WINDOW_WIDTH - 1) + '\n') * (LOG_WINDOW_HEIGHT - 2)
        clear += ' ' * (LOG_WINDOW_WIDTH-1)
        self.log_window.addstr(clear)

        if self.log_curr_line >= 0:
            log_dsp_line = self.log_curr_line
        else:
            log_dsp_line = len(self.log_content) - 1

        log_dsp_buffer = []
        while log_dsp_line >= 0:
            log_line_buffer = self.log_parse_oneline(self.log_content[log_dsp_line][1])
            log_dsp_buffer = log_line_buffer + log_dsp_buffer
            log_dsp_line -= 1
            if len(log_dsp_buffer) >= (LOG_WINDOW_HEIGHT - 1):
                break
        remove_len = len(log_dsp_buffer) - (LOG_WINDOW_HEIGHT - 1)
        if remove_len > 0:
            log_dsp_buffer = log_dsp_buffer[remove_len:]

        self.curseslock.acquire()
        for i in range(len(log_dsp_buffer)):
            try:
                self.log_window.addstr(i+1, 0, log_dsp_buffer[i][1], curses.color_pair(log_dsp_buffer[i][0]))
            except:
                self.curseslock.release()
                if DEBUG:
                    print log_dsp_buffer[i]
                    print 'len:', len(log_dsp_buffer[i][1])
                    raise
                return

        self.log_window.refresh()
        self.cmd_window.refresh()
        self.curseslock.release()

    def device_list_display(self):
        self.curseslock.acquire()
        self.dev_window.move(1,0)
        clean = (' ' * (DEV_WINDOW_WIDTH-1) + '\n') * (LOG_WINDOW_HEIGHT - 2)
        clean += ' ' * (DEV_WINDOW_WIDTH-1)
        self.dev_window.addstr(clean)
        devices = list(self.device_list)
        devices.sort()
        if len(self.device_list) > 0:
            cuuid = devices[0].split(',')[0]
        linenum = 1
        for i in range(len(devices)):
            devfull = devices[i]
            dev = devfull.split(',')
            if dev[0] != cuuid:
                cuuid = dev[0]
                linenum += 1
            if dev[0] in self.alias_tuples:
                dev_str = str(i) + '.' + self.alias_tuples[dev[0]] + ':'
            else:
                dev_str = str(i) + '.' + dev[0] + ':'
            mid_len = DEV_WINDOW_WIDTH - len(dev_str) - 5
            devpath = dev[1].replace('/dev/','')
            if mid_len >= len(devpath):
                dev_str += devpath + ' '*(mid_len - len(devpath))
            else:
                dev_str += devpath[:mid_len]
            dev_str += '  ' + self.device_list[devfull]
            try:
                self.dev_window.addstr(linenum, 0, dev_str)
            except:
                pass
            linenum += 1
            if linenum >= LOG_WINDOW_HEIGHT:
                break
        self.dev_window.refresh()
        self.cmd_window.refresh()
        self.curseslock.release()

    def cmdrun_command_display(self, cmd, postion):
        self.curseslock.acquire()
        self.cmd_window.move(0, len('Command:'))
        self.cmd_window.addstr(' ' * (LOG_WINDOW_WIDTH + DEV_WINDOW_WIDTH - len('Command:')))
        self.cmd_window.move(0, len('Command:'))
        self.cmd_window.addstr(cmd)
        self.cmd_window.move(0, len('Command:') + postion)
        self.cmd_window.refresh()
        self.curseslock.release()

    def cmdrun_status_display(self, log):
        self.curseslock.acquire()
        self.cmd_window.move(1,0)
        self.cmd_window.addstr(' ' * (LOG_WINDOW_WIDTH + DEV_WINDOW_WIDTH))
        self.cmd_window.move(1,0)
        self.cmd_window.addstr(log[:(LOG_WINDOW_WIDTH + DEV_WINDOW_WIDTH)])
        self.cmd_window.refresh()
        self.curseslock.release()

    def packet_send_thread(self):
        heartbeat_timeout = time.time() + 10
        while self.keep_running:
            try:
                [type, content] = self.output_queue.get(block=True, timeout=0.1)
            except Queue.Empty:
                type = None
                pass
            if self.service_socket == None:
                continue
            if type == None:
                if time.time() < heartbeat_timeout:
                    continue
                heartbeat_timeout += 10
                data = pkt.construct(pkt.HEARTBEAT,'')
            else:
                data = pkt.construct(type, content)
            try:
                self.service_socket.send(data)
            except:
                self.connected = False
                continue

    def send_packet(self, type, content, timeout=0.1):
        if self.service_socket == None:
            return False
        try:
            self.output_queue.put([type, content], True, timeout)
            return True
        except Queue.Full:
            if DEBUG: self.log_display(time.time(), 'error: ouput buffer full, drop packet [{0] {1}]'.format(type, content))
        return False

    def get_devstr_by_index(self, index):
        devices = list(self.device_list)
        devices.sort()
        return devices[index]

    def get_index_by_devstr(self, devstr):
        devices = list(self.device_list)
        devices.sort()
        return devices.index(devstr)

    def login_and_get_server(self, controller_ip, controller_port):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if ENCRYPT:
            cert_file = path.join(path.dirname(path.abspath(__file__)), 'controller_certificate.pem')
            sock = ssl.wrap_socket(sock, cert_reqs=ssl.CERT_REQUIRED, ca_certs = cert_file)
        try:
            sock.connect((controller_ip, controller_port))
            sock.settimeout(3)
            msg = ''
        except:
            if DEBUG: traceback.print_exc()
            return None

        content = pkt.construct(pkt.ACCESS_LOGIN, 'terminal,' + self.uuid)
        try:
            sock.send(content)
        except:
            if DEBUG: traceback.print_exc()
            sock.close()
            return None

        try:
            data = sock.recv(MAX_MSG_LENGTH)
        except KeyboardInterrupt:
            sock.close()
            raise
        except:
            sock.close()
            return None

        if data == '':
            sock.close()
            return None

        type, length, value, data = pkt.parse(data)
        #print 'controller', type, value
        rets = value.split(',')
        if type != pkt.ACCESS_LOGIN or rets[0] != 'success':
            if rets[0] == 'invalid access key':
                self.log_display(time.time(), 'login to controller failed, invalid access key')
                accesskey_file = path.join(path.expanduser('~'), '.udterminal', '.accesskey')
                if path.exists(accesskey_file):
                    os.remove(accesskey_file)
            else:
                self.log_display(time.time(), 'login to controller failed, ret={0}'.format(value))
            sock.close()
            return None

        sock.close()
        return rets[1:]

    def connect_to_server(self, server_ip, server_port, certificate):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if certificate != 'None':
            try:
                certfile = path.join(path.expanduser('~'), '.udterminal', 'certificate.pem')
                f = open(certfile, 'wt')
                f.write(certificate)
                f.close()
                sock = ssl.wrap_socket(sock, cert_reqs=ssl.CERT_REQUIRED, ca_certs=certfile)
            except:
                return 'fail'
        try:
            sock.connect((server_ip, server_port))
            while self.output_queue.empty() == False:
                self.output_queue.get()
            self.service_socket = sock
            self.connected = True
            return 'success'
        except:
            return 'fail'

    def server_interaction(self):
        self.connected = False
        self.service_socket = None
        msg = ''
        while self.keep_running:
            try:
                if self.connected == False:
                    raise ConnectionLost

                new_msg = self.service_socket.recv(MAX_MSG_LENGTH)
                if new_msg == '':
                    raise ConnectionLost
                    break

                msg += new_msg
                while msg != '':
                    type, length, value, msg = pkt.parse(msg)
                    if type == pkt.TYPE_NONE:
                        break

                    #print type, length
                    if type == pkt.FILE_BEGIN:
                        if 'file' in locals() and file.closed == False:
                            file.close()
                        filename = path.join(path.expanduser('~'), '.udterminal', value)
                        file = open(filename, 'w')
                        continue
                    if type == pkt.FILE_DATA:
                        if 'file' in locals() and file.closed == False:
                            file.write(value)
                        continue
                    if type == pkt.FILE_END:
                        if 'file' in locals():
                            file.close()
                        continue
                    if type == pkt.ALL_DEV:
                        new_list = {}
                        clients = value.split(':')
                        for c in clients:
                            if c == '':
                                continue
                            devs = c.split(',')
                            uuid = devs[0]
                            devs = devs[1:]
                            for d in devs:
                                if d == '':
                                    continue
                                [dev, using] = d.split('|')
                                new_list[uuid + ',' + dev] = using

                        for dev in list(new_list):
                            self.device_list[dev] = new_list[dev]

                        for dev in list(self.device_list):
                            if dev not in list(new_list):
                                self.device_list.pop(dev)
                        self.device_list_display()
                        continue
                    if type == pkt.DEVICE_LOG:
                        dev = value.split(':')[0]
                        logtime = value.split(':')[1]
                        log = value[len(dev) + 1 + len(logtime):]
                        try:
                            logtime = float(logtime)
                        except:
                            continue
                        if dev not in list(self.device_list):
                            continue
                        index = self.get_index_by_devstr(dev)
                        if dev in self.log_subscribed:
                            log =  str(index) + log
                            self.log_display(logtime, log)
                        continue
                    if type == pkt.DEVICE_DEBUG_DATA:
                        dev = value.split(':')[0]
                        if dev not in list(self.device_list):
                            continue
                        data = value[len(dev)+1:]
                        if dev not in self.debug_socks:
                            continue
                        try:
                            self.debug_socks[dev].sendall(data)
                        except:
                            index = self.get_index_by_devstr(dev)
                            self.log_display(time.time(), 'forwording debug data for {0} failed'.format(index))
                    if type in [pkt.DEVICE_DEBUG_START, pkt.DEVICE_DEBUG_REINIT, pkt.DEVICE_DEBUG_STOP]:
                        #self.log_display(time.time(), '{0}'.format([type, value]))
                        try:
                            [dev, result] = value.split(':')
                        except:
                            self.log_display(time.time(), 'error: wrong data {0} {1}'.format(type,value))
                            continue
                        if dev not in list(self.device_list):
                            continue
                        message = [type, dev, result]
                        if self.debug_queue.full() == True:
                            self.log_display(time.time(), 'local debug deamon busy, discard packet {0}'.format([type, value]))
                            continue
                        self.debug_queue.put(message)
                        continue
                    if type == pkt.CMD_DONE:
                        self.cmd_excute_return = value
                        self.cmd_excute_state = 'done'
                        self.cmd_excute_event.set()
                        continue
                    if type == pkt.CMD_ERROR:
                        self.cmd_excute_return = value
                        self.cmd_excute_state = 'error'
                        self.cmd_excute_event.set()
                        continue
                    if type == pkt.TERMINAL_LOGIN:
                        if value == 'success':
                            continue
                        self.cmdrun_status_display('login to server failed, retry later ...')
                        time.sleep(10)
                        continue
            except ConnectionLost:
                self.connected = False
                if self.service_socket != None:
                    self.service_socket.close()
                    self.service_socket = None
                    self.cmdrun_status_display('connection to server lost, try reconnecting...')

                result = self.login_and_get_server(self.controller_ip, self.controller_port)
                if result == None:
                    self.cmdrun_status_display('login to controller failed, retry later...')
                    time.sleep(5)
                    continue
                try:
                    [server_ip, server_port, token, certificate] = result
                    server_port = int(server_port)
                except:
                    self.log_display(time.time(), 'login to controller failed, invalid return={0}'.format(result))
                    self.cmdrun_status_display('login to controller failed, retry later...')
                    time.sleep(5)
                    continue

                result = self.connect_to_server(server_ip, server_port, certificate)
                if result == 'success':
                    self.cmdrun_status_display('connect to server succeeded')
                else:
                    self.cmdrun_status_display('connect to server failed, retry later ...')
                    time.sleep(5)
                    continue

                data = self.uuid + ',' + token
                self.send_packet(pkt.TERMINAL_LOGIN, data)
            except:
                if DEBUG:
                    raise
                break
        self.keep_running = False;

    def debug_interaction(self):
        debug_sessions = {}
        while self.keep_running:
            if debug_sessions == {} or self.debug_queue.empty() == False:
                [type, device, result] = self.debug_queue.get()
                index = self.get_index_by_devstr(device)
                if type == pkt.DEVICE_DEBUG_START:
                    if result != 'success':
                        self.log_display(time.time(), 'error: start debug device {0} failed, ret={1}'.format(index, result))
                        self.cmd_excute_return = result
                        self.cmd_excute_state = 'error'
                        self.cmd_excute_event.set()
                        continue

                    if device in debug_sessions:
                        self.cmd_excute_return = port
                        self.cmd_excute_state = 'done'
                        self.cmd_excute_event.set()
                        port = debug_sessions[device]['port']
                        self.log_display(time.time(), 'start debugging {0} succeed, serve at port {1}'.format(index, port))
                        continue
                    used_ports = []
                    for device in debug_sessions:
                        used_ports.append(debug_sessions[device]['port'])
                    port = 4120 + ord(os.urandom(1)) * ord(os.urandom(1)) / 64
                    while port in used_ports:
                        port = 4120 + ord(os.urandom(1)) * ord(os.urandom(1)) / 64
                    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    sock.setblocking(False)
                    try:
                        sock.bind(('localhost', port))
                    except:
                        self.log_display(time.time(), 'start debugging {0} failed, listen at port {1} failed'.format(index, port))
                        self.cmd_excute_return = 'fail'
                        self.cmd_excute_state = 'error'
                        self.cmd_excute_event.set()
                        continue
                    sock.listen(1)
                    debug_sessions[device] = {'port': port, 'listen_sock': sock}
                    self.log_display(time.time(), 'start debugging {0} succeed, serve at port {1}'.format(index, port))
                    self.cmd_excute_return = port
                    self.cmd_excute_state = 'done'
                    self.cmd_excute_event.set()
                elif type == pkt.DEVICE_DEBUG_REINIT:
                    if result == 'success':
                        continue
                    self.log_display(time.time(), 'error: restart debugging device {0} failed, ret={1}, closing debug session'.format(index, result))
                    if device not in debug_sessions:
                        continue
                    if 'serve_sock' in debug_sessions[device]:
                        self.debug_socks.pop(device)
                        debug_sessions[device]['serve_sock'].close()
                    debug_sessions[device]['listen_sock'].close()
                    debug_sessions.pop(device)
                elif type == pkt.DEVICE_DEBUG_STOP:
                    self.cmd_excute_return = result
                    if result == 'success':
                        self.cmd_excute_state = 'done'
                    else:
                        self.cmd_excute_state = 'error'
                    self.cmd_excute_event.set()

                    if device not in debug_sessions:
                        continue
                    self.log_display(time.time(), 'stop debugging {0}: {1}'.format(index, result))
                    if 'serve_sock' in debug_sessions[device]:
                        debug_sessions[device]['serve_sock'].close()
                    debug_sessions[device]['listen_sock'].close()
                    debug_sessions.pop(device)
                else:
                    self.log_display(time.time(), "error: wrong debug msg type '{0}'".format(type))
                continue

            select_list = {}
            for device in debug_sessions:
                if 'serve_sock' in debug_sessions[device]:
                    sock = debug_sessions[device]['serve_sock']
                    operation = 'recv'
                else:
                    sock = debug_sessions[device]['listen_sock']
                    operation = 'accept'
                select_list[sock] = [operation, device]

            r, w, e = select.select(list(select_list), [], [], 0.02)
            for sock in r:
                [operation, device] = select_list[sock]
                if operation == 'accept':
                    conn, addr = sock.accept()
                    conn.setblocking(0)
                    debug_sessions[device]['serve_sock'] = conn
                    self.debug_socks[device] = conn
                else:
                    try:
                        data = sock.recv(MAX_MSG_LENGTH)
                    except:
                        data = None
                    if not data:
                        debug_sessions[device].pop('serve_sock')
                        self.debug_socks.pop(device)
                        sock.close()
                        #self.log_display(time.time(), 'debug client disconnected')
                        self.send_packet(pkt.DEVICE_DEBUG_REINIT, device)
                        continue
                    content = device + ':' + data
                    self.send_packet(pkt.DEVICE_DEBUG_DATA, content)

    def parse_device_index(self, index_str):
        max = len(self.device_list) - 1
        if '-' in index_str:
            index_split = index_str.split('-')
            if len(index_split) != 2 or index_split[0].isdigit() == False or index_split[1].isdigit() == False:
                return [];
            try:
                start = int(index_split[0])
                end   = int(index_split[1])
                if start > max and end > max:
                    return []
                if start > max:
                    start = max
                if end > max:
                    end = max
            except:
                return [];
            if start < end:
                return range(start, end + 1, 1)
            else:
                return range(start, end - 1, -1)
        else:
            if index_str.isdigit() == False:
                return []
            try:
                index = int(index_str)
            except:
                return []
            if index >= len(self.device_list):
                return []
            return [index]

    def wait_cmd_excute_done(self, timeout):
        self.cmd_excute_state = 'wait_response'
        self.cmd_excute_return = None
        self.cmd_excute_event.clear()
        if self.cmd_excute_event.wait(timeout) == False:
            self.cmd_excute_state = 'timeout'

    def send_file_to_client(self, filename, index):
        status_str = 'sending file {0} to {1}...'.format(filename, self.get_devstr_by_index(index).split(',')[0])
        self.cmdrun_status_display(status_str)
        filehash = pkt.hash_of_file(filename)
        devstr = self.get_devstr_by_index(index)

        #send file begin
        content = devstr  + ':' + filehash + ':' + filename.split('/')[-1]
        retry = 4
        while retry > 0:
            self.send_packet(pkt.FILE_BEGIN, content)
            self.wait_cmd_excute_done(0.2)
            if self.cmd_excute_state == 'timeout':
                retry -= 1;
                continue
            if self.cmd_excute_return == 'busy':
                time.sleep(5)
                continue
            elif self.cmd_excute_return == 'ok' or self.cmd_excute_return == 'exist':
                break
            else:
                status_str += 'error'
                self.cmdrun_status_display(status_str)
                return False
        if retry == 0:
            status_str += 'error'
            self.cmdrun_status_display(status_str)
            return False

        if self.cmd_excute_return == 'exist':
            status_str += 'done'
            self.cmdrun_status_display(status_str)
            return True

        #send file data
        seq = 0
        file = open(filename,'r')
        header = devstr  + ':' + filehash + ':' + str(seq) + ':'
        content = file.read(8192)
        while(content):
            retry = 4
            while retry > 0:
                self.send_packet(pkt.FILE_DATA, header + content)
                self.wait_cmd_excute_done(0.2)
                if self.cmd_excute_return == None:
                    retry -= 1;
                    continue
                elif self.cmd_excute_return != 'ok':
                    status_str += 'error'
                    self.cmdrun_status_display(status_str)
                    file.close()
                    return False
                break

            if retry == 0:
                status_str += 'error'
                self.cmdrun_status_display(status_str)
                file.close()
                return False

            seq += 1
            header = devstr  + ':' + filehash + ':' + str(seq) + ':'
            content = file.read(8192)
        file.close()

        #send file end
        content = devstr  + ':' + filehash + ':' + filename.split('/')[-1]
        retry = 4
        while retry > 0:
            self.send_packet(pkt.FILE_END, content)
            self.wait_cmd_excute_done(0.2)
            if self.cmd_excute_return == None:
                retry -= 1;
                continue
            elif self.cmd_excute_return != 'ok':
                status_str += 'error'
                self.cmdrun_status_display(status_str)
                return False
            break
        if retry == 0:
            status_str += 'error'
            self.cmdrun_status_display(status_str)
            return False

        status_str += 'done'
        self.cmdrun_status_display(status_str)
        return True

    def erase_devices(self, args):
        devs = args
        if devs == []:
            self.cmdrun_status_display('Usage error: please specify devices you want to programg')
            return False
        succeed = []; failed = []
        for dev in devs:
            indexes = self.parse_device_index(dev)
            if indexes == []:
                self.cmdrun_status_display('invalid device index {0}'.format(dev))
                continue

            for index in indexes:
                status_str = 'erasing {0}.{1}...'.format(index, self.get_devstr_by_index(index))
                self.cmdrun_status_display(status_str)
                dev_str = self.get_devstr_by_index(index)
                self.send_packet(pkt.DEVICE_ERASE, dev_str);
                self.wait_cmd_excute_done(10)
                status_str += self.cmd_excute_state
                self.cmdrun_status_display(status_str)
                if self.cmd_excute_state == 'done':
                    succeed.append(index)
                else:
                    failed.append(index)
                self.cmd_excute_state = 'idle'
                if dev_str not in self.using_list:
                    self.using_list.append(dev_str)
        status_str = ''
        if succeed != []:
            status_str += 'succeed: {0}'.format(succeed)
        if failed != []:
            if status_str != '':
                status_str += ', '
            status_str += 'failed: {0}'.format(failed)
        self.cmdrun_status_display(status_str)

    def program_devices(self, args):
        if len(args) < 3:
            self.cmdrun_status_display('Usage: programg addresse filename device0 [device1 device2 ... deviceN]')
            return False
        if args[0].startswith('0x') == False:
            self.cmdrun_status_display('Usage error: wrong address input {0}, address should start with 0x'.format(args[0]))
            return False
        address  = args[0]
        filename = args[1]
        try:
            expandname = path.expanduser(filename)
        except:
            self.cmdrun_status_display('{0} does not exist'.format(filename))
            return False
        if path.exists(expandname) == False:
            self.cmdrun_status_display('{0} does not exist'.format(filename))
            return False
        filehash = pkt.hash_of_file(expandname)

        devs = args[2:]
        file_exist_at = []
        if devs == []:
            self.cmdrun_status_display('Usage error: please specify devices you want to programg')
            return False
        succeed = []; failed = []
        for dev in devs:
            indexes = self.parse_device_index(dev)
            if indexes == []:
                self.cmdrun_status_display('invalid device index {0}'.format(dev))
                continue
            for index in indexes:
                dev_str = self.get_devstr_by_index(index)
                [uuid, port] = dev_str.split(',')
                if uuid not in file_exist_at:
                    if self.send_file_to_client(expandname, index) == False:
                        failed.append(index)
                        continue
                    file_exist_at.append(uuid)
                status_str = 'programming {0} to {1}.{2}:{3}...'.format(filename, index, uuid, port.replace('/dev/', ''))
                self.cmdrun_status_display(status_str)
                content = dev_str + ',' + address + ',' + filehash
                self.send_packet(pkt.DEVICE_PROGRAM, content);
                self.wait_cmd_excute_done(270)
                status_str += self.cmd_excute_state
                self.cmdrun_status_display(status_str)
                if self.cmd_excute_state == 'done':
                    succeed.append(index)
                else:
                    failed.append(index)
                self.cmd_excute_state = 'idle'
                if dev_str not in self.using_list:
                    self.using_list.append(dev_str)
        status_str = ''
        if succeed != []:
            status_str += 'succeed: {0}'.format(succeed)
        if failed != []:
            if status_str != '':
                status_str += ', '
            status_str += 'failed: {0}'.format(failed)
        self.cmdrun_status_display(status_str)

    def control_devices(self, operate, args):
        if len(args) < 1:
            self.cmdrun_status_display('Usage error, usage: reset devices')
            return False

        operations = {'start':pkt.DEVICE_START, 'stop':pkt.DEVICE_STOP, 'reset':pkt.DEVICE_RESET}
        operate = operations[operate]
        succeed = []; failed = []
        for dev in args:
            indexes = self.parse_device_index(dev)
            if indexes == []:
                self.cmdrun_status_display('invalid device index {0}'.format(dev))
                return False

            for index in indexes:
                dev_str = self.get_devstr_by_index(index)
                self.send_packet(operate, dev_str)
                self.wait_cmd_excute_done(1.5)
                if self.cmd_excute_state == 'done':
                    succeed.append(index)
                else:
                    failed.append(index)
                self.cmd_excute_state = 'idle'
                if dev_str not in self.using_list:
                    self.using_list.append(dev_str)
        status_str = ''
        if succeed != []:
            status_str += 'succeed: {0}'.format(succeed)
        if failed != []:
            if status_str != '':
                status_str += ', '
            status_str += 'failed: {0}'.format(failed)
        self.cmdrun_status_display(status_str)

    def log_on_off(self, args):
        if len(args) < 2:
            self.cmdrun_status_display('Usage error, usage: log on/off devices')
            return False

        if args[0] == 'on':
            type = pkt.LOG_SUB
        elif args[0] == 'off':
            type = pkt.LOG_UNSUB
        else:
            self.cmdrun_status_display('Usage error, usage: log on/off devices')
            return False

        for dev in args[1:]:
            indexes = self.parse_device_index(dev)
            if not indexes:
                self.cmdrun_status_display('invalid device index {0}'.format(dev))
                continue
            for index in indexes:
                dev_str = self.get_devstr_by_index(index)
                if type == pkt.LOG_SUB and dev_str not in self.log_subscribed:
                    self.send_packet(type, dev_str)
                    self.log_subscribed.append(dev_str)
                elif type == pkt.LOG_UNSUB and dev_str in self.log_subscribed:
                    self.send_packet(type, dev_str)
                    self.log_subscribed.remove(dev_str)

    def log_download(self, args):
        if len(args) < 1:
            self.cmdrun_status_display('Usage error, usage: logdownload device0 [device1 ... deviceN]')
            return False

        succeed = []; failed = []
        for dev in args:
            indexes = self.parse_device_index(dev)
            if indexes == []:
                self.cmdrun_status_display('invalid device index {0}'.format(dev))
                return False
            for index in indexes:
                device = self.get_devstr_by_index(index).split(',')
                status_str = 'downloading log file for {0}.{1}:{2}...'.format(index, device[0], device[1].replace('/dev/', ''))
                self.cmdrun_status_display(status_str)
                content = ','.join(device)
                self.send_packet(pkt.LOG_DOWNLOAD, content)
                self.wait_cmd_excute_done(480)
                if self.cmd_excute_state == 'done':
                    succeed.append(index)
                else:
                    failed.append(index)
                self.cmd_excute_state = 'idle'
        status_str = ''
        if succeed != []:
            status_str += 'succeed: {0}'.format(succeed)
        if failed != []:
            if status_str != '':
                status_str += ', '
            status_str += 'failed: {0}'.format(failed)
        self.cmdrun_status_display(status_str)
        return (len(failed) == 0)

    def run_command(self, args, uselast = False):
        if uselast == False:
            if len(args) < 2:
                self.cmdrun_status_display('Usage error, usage: runcmd device cmd_arg0 [cmd_arg1 cmd_arg2 ... cmd_argN]')
                return False

            indexes = self.parse_device_index(args[0])
            if indexes == []:
                self.cmdrun_status_display('invalid device index {0}'.format(args[0]))
                return False
            args = args[1:]
            self.last_runcmd_dev = self.get_devstr_by_index(indexes[-1])
        else:
            if self.last_runcmd_dev == []:
                self.cmdrun_status_display('Error: you have not excute any remote command with runcmd yet, no target remembered')
                return False

            if len(args) < 1:
                self.cmdrun_status_display('Usage error, usage: !cmd_arg0 [cmd_arg1 cmd_arg2 ... cmd_argN]')
                return False

            if self.last_runcmd_dev not in list(self.device_list):
                self.cmdrun_status_display('Error: remembered target no longer exists')
                return False
            indexes = [self.get_index_by_devstr(self.last_runcmd_dev)]

        succeed = []; failed = []
        for index in indexes:
            dev_str = self.get_devstr_by_index(index)
            content = dev_str + ':' + '|'.join(args)
            self.send_packet(pkt.DEVICE_CMD, content)
            self.wait_cmd_excute_done(1.5)
            if self.cmd_excute_state == 'done':
                succeed.append(index)
            else:
                failed.append(index)
            self.cmd_excute_state = 'idle'
            if self.cmd_excute_return == None:
                self.cmd_excute_return = 'timeout'
            status_str = '{0} run: '.format(index) + ' '.join(args) + ', ' + self.cmd_excute_return
            self.cmdrun_status_display(status_str)
            if dev_str not in self.using_list:
                self.using_list.append(dev_str)
        if len(indexes) == 1:
            return True
        status_str = ''
        if succeed != []:
            status_str += 'succeed: {0}'.format(succeed)
        if failed != []:
            if status_str != '':
                status_str += ', '
            status_str += 'failed: {0}'.format(failed)
        self.cmdrun_status_display(status_str)
        return (len(failed) == 0)

    def run_debug(self, args, uselast = False):
        if len(args) < 2 or args[0] not in ['start', 'stop']:
            self.cmdrun_status_display('Usage error, usage: debug start/stop device1 [device2 .. deviceN] ')
            return False
        operation = args[0]
        devs = args[1:]
        indexes = []; devices = {}
        for dev in devs:
            indexes += self.parse_device_index(dev)
        if indexes == []:
            self.cmdrun_status_display('invalid device index {0}'.format(' '.join(devs)))
            return False
        indexes = list(set(indexes))
        for index in indexes:
            devices[self.get_devstr_by_index(index)] = index

        if operation == 'start':
            type = pkt.DEVICE_DEBUG_START
        else:
            type = pkt.DEVICE_DEBUG_STOP

        succeed = []; failed = []
        for device in list(devices):
            if device not in self.device_list:
                continue
            self.send_packet(type, device)
            self.wait_cmd_excute_done(2)
            if self.cmd_excute_state == 'done':
                succeed.append(devices[device])
            else:
                failed.append(devices[device])
        status_str = ''
        if succeed != []:
            status_str += 'succeed: {0}'.format(succeed)
        if failed != []:
            if status_str != '':
                status_str += ', '
            status_str += 'failed: {0}'.format(failed)
        self.cmdrun_status_display(status_str)
        return (len(failed) == 0)

    def client_alias(self, args):
        if len(args) < 1:
            self.cmdrun_status_display('Usage error, usage: alias id0:name0 [id1:name1 ... idN:nameN]')
            return False

        for arg in args:
            alias = arg.split(':')
            if len(alias) != 2:
                self.cmdrun_status_display('Usage error, unrecongnized alias tuple {0}'.format(arg))
                continue
            self.alias_tuples[alias[0]] = alias[1]
        try:
            alias_file = path.join(path.expanduser('~'), '.udterminal', '.alias')
            file = open(alias_file, 'wb')
            json.dump(self.alias_tuples, file)
            file.close()
        except:
            self.cmdrun_status_display('error: save alias record failed')
        self.device_list_display()

    def print_help_info(self):
        self.log_display(time.time(), 'Supported commands and usage examples:')
        self.log_display(time.time(), ' 1.erase      [er]: erase flash of devices')
        self.log_display(time.time(), '           example: erase 0 1 2-5')
        self.log_display(time.time(), ' 2.program    [pg]: program fireware to devices')
        self.log_display(time.time(), '           example: program 0x40000 aos_esp32.bin 0 1 5-10')
        self.log_display(time.time(), ' 3.reset      [rs]: reset/restart devices')
        self.log_display(time.time(), '           example: reset 0 1 3-8')
        self.log_display(time.time(), ' 4.stop       [sp]: stop devices')
        self.log_display(time.time(), '           example: stop 0 5-7 9')
        self.log_display(time.time(), ' 5.start      [st]: start devices')
        self.log_display(time.time(), '           example: start 3-5 0 1')
        self.log_display(time.time(), ' 6.runcmd     [rc]: run command at remote device')
        self.log_display(time.time(), '           example: runcmd 0 ping fc:00:00:10:11:22:33:44')
        self.log_display(time.time(), '                    runcmd 0-10 umesh status')
        self.log_display(time.time(), ' 7.^              : run command at latest (runcmd) remote device')
        self.log_display(time.time(), '           example: ^ping fc:00:00:10:11:22:33:44')
        self.log_display(time.time(), ' 8.debug      [db]: start/stop debugging remote device')
        self.log_display(time.time(), '           example: debug start 0')
        self.log_display(time.time(), '                    debug stop 1')
        self.log_display(time.time(), ' 9.log        [lg]: turn on/off log display for devices, eg.: log on 1')
        self.log_display(time.time(), '           example: log on 1 2 5-8; log off 2-5 7')
        self.log_display(time.time(), ' 10.logdownload[ld]: download log file of device from server')
        self.log_display(time.time(), '           example: logdownload 0-2 5')
        self.log_display(time.time(), ' 11.alias     [al]: alias names to client ids')
        self.log_display(time.time(), '           example: alias 1234567890123456:Pi1@HZ')
        self.log_display(time.time(), ' 12.quit       [q]: quit this terminal')
        self.log_display(time.time(), ' 13.help       [h]: print help infomation')
        self.log_display(time.time(), ' ')
        self.log_display(time.time(), 'Supported command editing shortcuts:')
        self.log_display(time.time(), ' Ctrl + a - go to the start of the command line')
        self.log_display(time.time(), ' Ctrl + e - go to the end of the command line')
        self.log_display(time.time(), ' Ctrl + k - delete from cursor to the end of the command line')
        self.log_display(time.time(), ' Ctrl + u - delete from cursor to the start of the command line')

    def process_cmd(self, cmd):
        if self.connected == False:
            self.cmdrun_status_display("process command '{0}' failed, connection to server lost".format(cmd))
        cmd_argv = cmd.split(' ')
        self.cmdrun_status_display('')
        cmd = cmd_argv[0]; args = cmd_argv[1:]
        if cmd == 'erase' or cmd == 'er':
            self.erase_devices(args)
        elif cmd == 'program' or cmd == 'pg':
            self.program_devices(args)
        elif cmd == 'reset' or cmd == 'rs':
            self.control_devices('reset', args)
        elif cmd == 'start' or cmd == 'st':
            self.control_devices('start', args)
        elif cmd == 'stop' or cmd == 'sp':
            self.control_devices('stop', args)
        elif cmd == 'log' or cmd == 'lg':
            self.log_on_off(args)
        elif cmd == 'logdownload' or cmd == 'ld':
            self.log_download(args)
        elif cmd == 'runcmd' or cmd == 'rc':
            self.run_command(args, uselast=False)
        elif cmd == 'debug' or cmd == 'db':
            self.run_debug(args)
        elif cmd[0] == '^':
            self.run_command([cmd[1:]] + args, uselast=True)
        elif cmd == 'alias' or cmd == 'al':
            self.client_alias(args)
        elif cmd == 'help' or cmd == 'h':
            self.print_help_info()
        else:
            self.cmdrun_status_display('unknown command:' + cmd)

    def user_interaction(self):
        cmd = ''
        saved_cmd = ''
        history_index = -1
        p = 0
        while self.keep_running:
            c = self.cmd_window.getch()
            if c == ord('\n'):
                if self.log_curr_line != -1:
                    self.log_curr_line = -1
                    self.log_display(time.time(), '')
                if cmd == 'quit' or cmd == 'q' :
                    self.keep_running = False
                    time.sleep(0.2)
                    break
                elif cmd != '':
                    self.process_cmd(cmd)
                    self.cmd_history = [cmd] + self.cmd_history
                cmd = ''
                saved_cmd = ''
                history_index = -1
                p = 0
                self.cmdrun_command_display(cmd, 0)
            elif c == curses.KEY_BACKSPACE or c == 127 or c == 8: #DELETE
                if cmd[0:p] == '':
                    continue
                newcmd = cmd[0:p-1] + cmd[p:]
                cmd = newcmd
                p -= 1
                self.cmdrun_command_display(cmd, p)
            elif c == 259 and len(self.cmd_history) > 0: #KEY_UP
                if history_index == -1:
                    saved_cmd = cmd
                if history_index < (len(self.cmd_history) - 1):
                    history_index += 1
                cmd = self.cmd_history[history_index]
                p = len(cmd)
                self.cmdrun_command_display(cmd, p)
            elif c == 258 and len(self.cmd_history) > 0: #KEY_DOWN
                if history_index <= -1:
                    history_index = -1
                    continue
                history_index -= 1
                if history_index >= 0:
                    cmd = self.cmd_history[history_index]
                else:
                    cmd = saved_cmd
                p = len(cmd)
                self.cmdrun_command_display(cmd, p)
            elif c == 260: #KEY_LEFT
                if p > 0:
                    p -= 1
                    self.cmdrun_command_display(cmd, p)
                continue
            elif c == 261: #KEY_RIGHT
                if p < len(cmd):
                    p += 1
                    self.cmdrun_command_display(cmd, p)
                continue
            elif c == 1: #ctrl-a: go to the start of the command line
                p = 0
                self.cmdrun_command_display(cmd, p)
                continue
            elif c == 3: #ctrl-c: quit
                self.keep_running = False
                time.sleep(0.2)
                break
            elif c == 5: #ctrl-e: go to the end of the command line
                p = len(cmd)
                self.cmdrun_command_display(cmd, p)
                continue
            elif c == 11: #ctrl-k: delete from cursor to the end of the command line
                cmd = cmd[0:p]
                self.cmdrun_command_display(cmd, p)
                continue
            elif c == 21: #ctrl-u: delete from cursor to the start of the command line
                cmd = cmd[p:]
                p = 0
                self.cmdrun_command_display(cmd, p)
                continue
            elif c == curses.KEY_MOUSE:
                mouse_state = curses.getmouse()[4]
                if mouse_state == MOUSE_SCROLL_UP:
                    if self.log_curr_line == -1:
                        self.log_curr_line = len(self.log_content) - 2
                    elif self.log_curr_line > 0:
                        self.log_curr_line -= 1
                    self.log_display(time.time(), '')
                    continue
                elif mouse_state == MOUSE_SCROLL_DOWN:
                    if self.log_curr_line == -1:
                        continue

                    self.log_curr_line += 1
                    if self.log_curr_line >= len(self.log_content):
                        self.log_curr_line = -1
                    self.log_display(time.time(), '')
                    continue
            elif c == curses.KEY_RESIZE:
                try:
                    self.window_redraw()
                    self.device_list_display()
                    self.log_display(time.time(), '')
                    self.cmdrun_command_display(cmd, p);
                except:
                    if DEBUG:
                        raise
                    self.keep_running = False
                    break
            else:
                try:
                    c = str(unichr(c))
                    newcmd = cmd[0:p] + c + cmd[p:]
                    cmd = newcmd
                    p += 1
                except:
                    self.cmdrun_status_display('Error: unsupported unicode character {0}'.format(c))
                    continue
                self.cmdrun_command_display(cmd, p)

    def run(self, controller_ip, controller_port):
        #connect to server
        self.controller_ip = controller_ip
        self.controller_port = controller_port
        thread.start_new_thread(self.packet_send_thread, ())
        thread.start_new_thread(self.server_interaction, ())
        thread.start_new_thread(self.debug_interaction, ())
        while self.keep_running:
            try:
                self.user_interaction()
            except KeyboardInterrupt:
                self.keep_running = False
                time.sleep(0.2)
            except:
                if DEBUG:
                    raise

    def deinit(self):
        curses.nocbreak()
        self.cmd_window.keypad(0)
        curses.echo()
        curses.endwin()
        if self.service_socket:
            self.service_socket.close()
        try:
            if len(self.cmd_history) > 0:
                cmd_history_file = path.join(path.expanduser('~'), '.udterminal', '.cmd_history')
                file = open(cmd_history_file, 'wb')
                json.dump(self.cmd_history, file)
                file.close()
        except:
            self.cmdrun_status_display('error: save command history failed')

    def terminal_func(self, controller_ip, controller_port):
        ret = self.init()
        if ret == 'read access key fail':
            print ''
            return
        if ret == 'success':
            self.run(controller_ip, controller_port)
        if ret == 'UI failed':
            print 'initilize UI window failed, try increase your terminal window size first'
        self.deinit()



#!/usr/bin/env python

# Copyright (c) 2016 aos Limited, All Rights Reserved
# SPDX-License-Identifier: Apache-2.0

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.

# You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied.

from __future__ import print_function
import argparse
import json
import filecmp
import traceback
import glob
import zipfile
import platform
import serial

from .util import *
from .repo import *
from serial.tools import miniterm
from serial.tools.list_ports import comports

def get_host_os():
    host_os = platform.system()
    if host_os == 'Windows':
        host_os = 'Win32'
    elif host_os == 'Linux':
        if platform.machine().endswith('64'):
            bit = '64'
        else:
            bit = '32'
        host_os += bit
    elif host_os == 'Darwin':
        host_os = 'OSX'
    else:
        host_os = None
    return host_os

def cd_aos_root():
    original_dir = os.getcwd()
    host_os = get_host_os()
    if host_os == 'Win32':
        sys_root = re.compile(r'^[A-Z]{1}:\\$')
    else:
        sys_root = re.compile('^/$')
    while os.path.isdir('./include/aos') == False:
        os.chdir('../')
        if sys_root.match(os.getcwd()):
            return 'fail', original_dir
    return 'success', original_dir

def which(program):
    if platform.system() == 'Windows' and program.endswith('.exe') == False:
        program += '.exe'

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None

def cmd_version_match(cmd, version):
    cmd = which(cmd)
    if cmd == None:
        return False
    if version == 'all':
        return True
    match = False
    ret = subprocess.check_output([cmd, '-v'], stderr=subprocess.STDOUT)
    lines = ret.split('\n')
    for line in lines:
        if version in line:
            match = True
            break;
    return match

def download_toolchains(downloads):
    tmp_dir = 'tmp_{0:02x}'.format(ord(os.urandom(1)))
    while os.path.isdir(tmp_dir):
        tmp_dir = 'tmp_{0:02x}'.format(ord(os.urandom(1)))
    try:
        os.mkdir(tmp_dir)
    except:
        print('download toolchains failed: unable to create a temp folder')
        return
    try:
        os.chdir(tmp_dir)
        for download in downloads:
            result = 0
            name, git_url, dst_dir = download
            print('toolchain {} missing, start download ...'.format(name))
            print(git_url + ' -> ' + dst_dir)
            result += subprocess.call(['git', 'clone', '--depth=1', git_url, name])
            if result > 0:
                print('git clone toolchain {} failed'.format(name))
                print('You can mannually try fix this problem by running:')
                print('    git clone {} {}'.format(git_url, name))
                print('    mv {0}/main {1} && rm -rf {0}'.format(name, dst_dir))
            src_dir = name + '/main'
            if os.path.exists(src_dir) == False:
                print('toolchain folder {} none exist'.format(src_dir))
                result += 1
            if result == 0:
                dst_dir = '../' + dst_dir
                if os.path.isfile(dst_dir) == True:
                    os.remove(dst_dir)
                if os.path.isdir(dst_dir) == True:
                    shutil.rmtree(dst_dir)
                if os.path.isdir(os.path.dirname(dst_dir)) == False:
                    os.makedirs(os.path.dirname(dst_dir))
                shutil.move(src_dir, dst_dir)
                print('download toolchain {} succeed'.format(name))
            else:
                print('download toolchain {} failed'.format(name))
    except:
        traceback.print_exc()
    finally:
        os.chdir('../')
    try:
        shutil.rmtree(tmp_dir)
    except:
        print("toolchain auto-install error: can not remove temp folder {}, please remove it manually".format(tmp_dir))
        pass

def _install_toolchains(build_args):
    board = None
    host_os = get_host_os()
    if host_os == None:
        error('Unsupported Operating System!')

    #cd to aos root_dir
    ret, original_dir = cd_aos_root()
    if ret != 'success':
        error("not in AliOS-Things source code directory")

    #check config file to be enable this function (for backward compatability)
    autodownload_enable = False
    if os.path.exists('build/toolchain_autodownload.config'):
        try:
            with open('build/toolchain_autodownload.config') as file:
                if 'yes' in file.read(): autodownload_enable = True
        except:
            pass
    if autodownload_enable == False:
        if os.path.isdir(original_dir): os.chdir(original_dir)
        return

    for arg in build_args:
        if '@' not in arg:
            continue
        args = arg.split('@')
        board = args[1]
        break

    if not board:
        if os.path.isdir(original_dir): os.chdir(original_dir)
        return
    if os.path.isdir(os.path.join('board', board)) == False:
        error('Can not find board {}'.format(board))

    downloads = []
    from .constant import boards
    if board in boards:
        print('Check if required tools for {} exist'.format(board))
        for toolchain in boards[board]:
            name = toolchain['name']
            command = toolchain['command']
            version = toolchain['version']
            cmd_path = '{}/{}/bin/{}'.format(toolchain['path'], host_os, command)
            if cmd_version_match(cmd_path, version) == True:
                continue
            if toolchain['use_global'] and cmd_version_match(command, version) == True:
                continue
            git_url = toolchain['{}_url'.format(host_os)]
            if git_url == '':
                continue
            if is_domestic() is False:
                git_url = git_url.replace('gitee', 'github')
                git_url = git_url.replace('alios-things', 'aliosthings')
            downloads.append([name, git_url, '{}/{}'.format(toolchain['path'], host_os)])
    if len(downloads): download_toolchains(downloads)

    if os.path.isdir(original_dir): os.chdir(original_dir)

def _run_make(arg_list):
    #install dependent toolchains
    _install_toolchains(sys.argv[2:])

    # check operating system
    host_os = get_host_os()
    if host_os == None:
        error('Unsupported Operating System!')

    #cd to aos root_dir
    ret, original_dir = cd_aos_root()
    if ret != 'success':
        error("not in AliOS-Things source code directory")

    make_cmds = {
        'Win32': 'cmd/win32/make.exe',
        'Linux64': 'cmd/linux64/make',
        'Linux32': 'cmd/linux32/make',
        'OSX': 'cmd/osx/make'
        }
    tools_dir = os.path.join(os.getcwd(), 'build').replace('\\', '/')
    make_cmd = os.path.join(tools_dir, make_cmds[host_os])

    # Run make command
    make_cmd_str = ' '.join([make_cmd, 'HOST_OS=' + host_os, 'TOOLS_ROOT=' + tools_dir] + list(arg_list))
    popen(make_cmd_str, shell=True, cwd=os.getcwd())
    if os.path.isdir(original_dir): os.chdir(original_dir)

def is_path_component(path):
    if not os.path.exists(path):
        return False

    for f in os.listdir(path):
        tmp_file = os.path.join(path, f)
        if os.path.isfile(tmp_file) and tmp_file.endswith('.mk') and os.path.basename(path) == f[:-3]:
            return True

    return False


def copy_directory(src_dir, dst_dir, sub_dir=False):
    if not os.path.exists(src_dir):
        return

    if os.path.exists(dst_dir):
        if sub_dir:
            dcmp = filecmp.dircmp(src_dir, dst_dir)
            if dcmp.left_only:
                order_dir = sorted(dcmp.left_only)
                for od in order_dir:
                    copy_directory(os.path.join(src_dir, od), os.path.join(dst_dir, od))
        return

    try:
        shutil.copytree(src_dir, dst_dir)
    except OSError as exc:  # python >2.5
        if exc.errno == errno.ENOTDIR:
            shutil.copy(src_dir, dst_dir)
        else:
            raise


def restore_aos_sdk(aos_sdk_path):
    with cd(aos_sdk_path):
        with open(os.devnull, 'w') as DEVNULL:
            try:
                if sys.platform.startswith('win'):
                    subprocess.Popen([git_cmd, 'checkout', '-f'], stdout=DEVNULL, stderr=DEVNULL)
                else:
                    subprocess.Popen(['nohup', git_cmd, 'checkout', '-f'], stdout=DEVNULL, stderr=DEVNULL)

            except ProcessException:
                pass


def generate_aos_program(program_path, aos_sdk_path, template):
    if not(program_path and os.path.isdir(program_path) and aos_sdk_path and os.path.isdir(aos_sdk_path)):
        error("AliOS-Things code isn't exits!!!")

    comp_name = os.path.basename(program_path)
    template = template if template else 'helloworld'
    pd = Program(program_path)

    example_dir = os.path.join(aos_sdk_path, 'example')
    src = os.listdir(os.path.join(example_dir, template))
    for s in src:
        if os.path.isdir(s):
            copy_directory(os.path.join(example_dir, template, s), os.path.join(program_path, s), True)
        else:
            shutil.copy(os.path.join(example_dir, template, s), program_path)

    with open(os.path.join(program_path, comp_name + '.mk'), 'w') as fout:
        with open(os.path.join(program_path, template + '.mk'), 'r') as fin:
            for line in fin:
                if line.startswith('NAME :='):
                    line = 'NAME := ' + comp_name
                fout.write(line)
    os.remove(os.path.join(program_path, template + '.mk'))

    pd.set_cfg(REMOTE_PATH, os.path.join(Global.get_path().replace(os.path.sep, '/'), 'remote'))
    pd.set_cfg(OS_PATH, aos_sdk_path.replace(os.path.sep, '/'))
    copy_directory(os.path.join(aos_sdk_path, '.vscode'), os.path.join(program_path, '.vscode'))
    vscode_cfg_file = os.path.join(program_path, '.vscode', 'settings.json')
    if os.path.exists(vscode_cfg_file):
        with open(vscode_cfg_file, 'r+') as f:
            vscode_cfg = json.load(f)
            vscode_cfg['aliosStudio.isAosSource'] = False
            f.seek(0)
            f.truncate()
            json.dump(vscode_cfg, f, indent = 4)


def get_path_root(path):
    tpath = os.path.dirname(path)
    while tpath != path:
        path = tpath
        tpath = os.path.dirname(path)
    return tpath


# Subparser handling
parser = argparse.ArgumentParser(prog='aos',
                                 description="Code management tool for aos - https://code.aliyun.com/aos/aos\nversion %s\n\nUse 'aos <command> -h|--help' for detailed help.\nOnline manual and guide available at https://code.aliyun.com/aos/aos-cube" % ver,
                                 formatter_class=argparse.RawTextHelpFormatter)
subparsers = parser.add_subparsers(title="Commands", metavar="           ")
parser.add_argument("--version", action="store_true", dest="version", help="print version number and exit")
subcommands = {}


# Process handling
def subcommand(name, *args, **kwargs):
    def __subcommand(command):
        if not kwargs.get('description') and kwargs.get('help'):
            kwargs['description'] = kwargs['help']
        if not kwargs.get('formatter_class'):
            kwargs['formatter_class'] = argparse.RawDescriptionHelpFormatter

        subparser = subparsers.add_parser(name, **kwargs)
        subcommands[name] = subparser

        for arg in args:
            arg = dict(arg)
            opt = arg['name']
            del arg['name']

            if isinstance(opt, basestring):
                subparser.add_argument(opt, **arg)
            else:
                subparser.add_argument(*opt, **arg)

        subparser.add_argument("-v", "--verbose", action="store_true", dest="verbose", help="Verbose diagnostic output")
        subparser.add_argument("-vv", "--very_verbose", action="store_true", dest="very_verbose",
                               help="Very verbose diagnostic output")

        def thunk(parsed_args):
            argv = [arg['dest'] if 'dest' in arg else arg['name'] for arg in args]
            argv = [(arg if isinstance(arg, basestring) else arg[-1]).strip('-').replace('-', '_')
                    for arg in argv]
            argv = {arg: vars(parsed_args)[arg] for arg in argv
                    if vars(parsed_args)[arg] is not None}

            return command(**argv)

        subparser.set_defaults(command=thunk)
        return command

    return __subcommand


# New command
@subcommand('new',
            dict(name='name', help='The program or component name'),
            dict(name=['-c', '--component'], action='store_true',
                 help='Force creation of an aos component. Default: auto.'),
            dict(name=['-t', '--template'], nargs='?',
                 help=''),
            dict(name='--scm', nargs='?',
                 help='Source control management. Currently supported: %s. Default: git' % ', '.join(
                     [s.name for s in scms.values()])),
            dict(name='--protocol', nargs='?',
                 help='Transport protocol when fetching the aos OS repository when creating new program. Supported: https, http, ssh, git. Default: inferred from URL.'),
            help='Create new aos program or component',
            description=(
                    "Create a new aos program (aos new helloworld) or component(aos new -c helloworld)\n"
                    "Supported source control management: git"))
def new(name, scm='git', component=False, template=None, protocol=None):
    global cwd_root

    d_type = 'component' if component else 'program'
    d_path = name if os.path.isabs(name) else os.path.abspath(name)

    if os.path.exists(d_path):
        p = Program(d_path)
        if (d_type == 'program' and not p.is_cwd) or (d_type == 'component' and Repo.isrepo(d_path)):
            error("A %s with name \"%s\" already exists." % (d_type, os.path.basename(d_path)), 1)

    if scm and scm != 'none':
        if os.path.isdir(d_path) and Repo.isrepo(d_path):
            repo = Repo.fromrepo(d_path)
            if repo.scm.name != scm:
                error("A repository already exists in \"%s\" based on %s. Please select a different name or location." %
                      (d_path, scm), 1)
        else:
            repo_scm = [s for s in scms.values() if s.name == scm.lower()]
            if not repo_scm:
                error(
                    "You have specified invalid source control management system\n"
                    "Please specify one of the following SCMs: %s" % ', '.join([s.name for s in scms.values()]), 1)
            repo_scm[0].init(d_path)

    if len(os.listdir(d_path)) > 1:
        warning("Directory \"%s\" is not empty." % d_path)

    action("Create new %s \"%s\" (%s)" % (d_type, d_path, scm))
    p = Program(d_path)
    if d_type == 'program':
        # This helps sub-commands to display relative paths to the created program
        # p.set_root()
        if is_domestic():
            aos_os_url = 'https://gitee.com/alios-things/AliOS-Things.git'
        else:
            aos_os_url = 'https://github.com/alibaba/AliOS-Things.git'
        url = aos_os_url + '#master'
        d = 'aos'
        aos_sdk_path = Global().get_cfg(AOS_SDK_PATH)
        if aos_sdk_path and os.path.isdir(os.path.join(aos_sdk_path, '.git')):
            action('AliOS-Things code has been download \"%s\"' % aos_sdk_path)
        else:
            try:
                change_path = Global().get_path()
                with cd(change_path):
                    add(url, None, protocol=protocol, top=False)
            except Exception as e:
                if os.path.isdir(os.path.join(d_path, d)):
                    rmtree_readonly(os.path.join(d_path, d))
                raise e

        generate_aos_program(d_path, Global().get_cfg(AOS_SDK_PATH), template)

        p.set_cfg(PROGRAM_PATH, d_path.replace(os.path.sep, '/'))
        p.set_cfg(PATH_TYPE, 'program')

    else:
        p.unset_root(d_path)
        with cd(d_path):
            with open(os.path.join(d_path, name + '.mk'), 'w') as fm:
                makefile_content = 'NAME := %s\n$(NAME)_SOURCES := %s\n$(NAME)_COMPONENTS :=\n$(NAME)_INCLUDES := .\n' % (
                    name, name + '.c')
                fm.write(makefile_content)

            with open(os.path.join(d_path, name + '.c'), 'w') as fc:
                fc.write('//source code file\n')

            with open(os.path.join(d_path, name + '.h'), 'w') as fh:
                fh.write('//head file\n')

    action("Create %s \"%s\" success" % (d_type, d_path))


def print_component_status(format):
    cwd_type = Repo().pathtype()
    os_name = OS_NAME

    if cwd_type == 'program':
        pd = Program(os.getcwd())
        aos_path = pd.get_cfg(OS_PATH)
        aos_remote_components = get_aos_components(pd.get_cfg(REMOTE_PATH))
        os_name = 'aos'
    else:
        aos_path = Global().get_cfg(AOS_SDK_PATH)

    aos_local_components = get_aos_components(aos_path)
    if format == 'json':
        json_components_dict = {}

        if cwd_type == 'program':
            pd = Program(os.getcwd())
            with cd(pd.get_cfg(PROGRAM_PATH)):
                cube_add_components = None
                cube_remove_components = None
                with open(CUBE_MAKEFILE, 'r') as fin:
                    for line in fin:
                        if line.startswith('CUBE_ADD_COMPONENTS :='):
                            cube_add_components = re.split('\s+', line[23:].strip())

                        if line.startswith('CUBE_REMOVE_COMPONENTS :='):
                            cube_remove_components = re.split('\s+', line[26:].strip())

                for path, name in aos_remote_components.items():
                    cube_add = False
                    cube_remove = False

                    rel_path = os.path.dirname(relpath(pd.get_cfg(PROGRAM_PATH), path))
                    if cube_add_components and os.path.dirname(rel_path) in cube_add_components:
                        cube_add = True
                    if cube_remove_components and os.path.dirname(rel_path) in cube_remove_components:
                        cube_remove = True

                    json_components_dict[rel_path] = {'name': name, 'cube_add': cube_add, 'cube_remove': cube_remove}

                for path, name in aos_local_components.items():
                    cube_add = False
                    cube_remove = False

                    rel_aos_path = os.path.dirname(relpath(aos_path, path))
                    rel_path = os.path.join(os_name, rel_aos_path).replace(os.path.sep, '/')

                    if cube_add_components and rel_aos_path in cube_add_components:
                        cube_add = True
                    if cube_remove_components and rel_aos_path in cube_remove_components:
                        cube_remove = True

                    json_components_dict[rel_path] = {'name': name, 'cube_add': cube_add, 'cube_remove': cube_remove}
        else:
            cube_add = False
            cube_remove = False
            for path, name in aos_local_components.items():
                json_components_dict[os.path.dirname(relpath(os.path.dirname(aos_path), path))] = {'name': name, 'cube_add': cube_add, 'cube_remove': cube_remove}

        print(json.dumps(json_components_dict, indent=4, sort_keys=True))
    else:
        if cwd_type != 'program':
            print('\nCurrent directory isn\'t AliOS-Things program, list AOS_SDK_PATH components.')
        print("\n                                                      AliOS-Things COMPONENTS                ")
        print('|===================================================================================================================|')

        print('| %-30s | %-80s |' % ('NAME', 'LOCATION'))

        if cwd_type == 'program':
            program_path = Program(os.getcwd()).get_cfg(PROGRAM_PATH)
            for path, name in aos_remote_components.items():
                print('| %-30s | %-80s |' % (name, os.path.dirname(relpath(program_path, path))))
        for path, name in aos_local_components.items():
            print('| %-30s | %-80s |' % (name, os.path.dirname(relpath(os.path.dirname(aos_path), path))))

        print('|===================================================================================================================|')


# List command
@subcommand('ls',
            dict(name=['-c', '--component'], action='store_true',
                 help='List component information.'),
            dict(name=['-f', '--format'], nargs='?',
                 help='Format output json, xml and so on.'),
            help='List cube info, default components info',
            description=("List the aos cube information."))
def list_(component=False, format=None):
    print_component_status(format)


# Import command
@subcommand('import',
            dict(name='url', help='URL of the program'),
            dict(name='path', nargs='?', help='Destination name or path. Default: current directory.'),
            dict(name=['-I', '--ignore'], action='store_true', help='Ignore errors related to cloning and updating.'),
            dict(name='--depth', nargs='?',
                 help='Number of revisions to fetch from the remote repository. Default: all revisions.'),
            dict(name='--protocol', nargs='?',
                 help='Transport protocol for the source control management. Supported: https, http, ssh, git. Default: inferred from URL.'),
            help='Import program from URL',
            description=(
                    "Imports aos program and its dependencies from a source control based URL\n"
                    "(GitHub, Bitbucket, aos.org) into the current directory or specified\npath.\n"
                    "Use 'aos add <URL>' to add a component into an existing program."))
def import_(url, path=None, ignore=False, depth=None, protocol=None, top=True):
    global cwd_root

    if not Repo.isurl(url) and not os.path.exists(url):
        url = AOS_COMPONENT_BASE_URL + '/' + url + '.git'
    repo = Repo.fromurl(url, path)
    if top:
        p = Program(path)
        if p and not p.is_cwd:
            error(
                "Cannot import program in the specified location \"%s\" because it's already part of a program \"%s\".\n"
                "Please change your working directory to a different location or use \"aos add\" to import the URL as a component." % (
                    os.path.abspath(repo.path), p.name), 1)

    protocol = Program().get_cfg('PROTOCOL', protocol)

    if os.path.isdir(repo.path) and len(os.listdir(repo.path)) > 1 and is_path_component(repo.path):
        info("Directory \"%s\" is a component." % repo.path, 1)
        return True

    if repo.clone(repo.url, repo.path, rev=repo.rev, depth=depth, protocol=protocol):
        with cd(repo.path):
            Program(repo.path).set_root()
            try:
                if repo.rev and repo.getrev() != repo.rev:
                    repo.checkout(repo.rev, True)
            except ProcessException as e:
                err = "Unable to update \"%s\" to %s" % (repo.name, repo.revtype(repo.rev, True))
                if depth:
                    err = err + (
                            "\nThe --depth option might prevent fetching the whole revision tree and checking out %s." % (
                    repo.revtype(repo.rev, True)))
                warning(err)
                return False
    else:
        err = "Unable to clone repository (%s)" % url
        warning(err)
        return False

    if os.path.isdir(repo.path):
        repo.sync()

        if top:  # This helps sub-commands to display relative paths to the imported program
            cwd_root = repo.path

        with cd(repo.path):
            deploy(ignore=ignore, depth=depth, protocol=protocol, top=False)

    return True


def get_component_dependencies(makefile):
    dependencies = []
    if makefile and makefile.endswith('.mk') and os.path.isfile(makefile):
        with open(makefile) as fin:
            for line in fin:
                line = line.strip('\n')
                while line.endswith('\\'):
                    line = line[:-1] + next(fin).rstrip('\n')
                if '$(NAME)_COMPONENTS' in line and not line.strip().startswith('#'):
                    line = line[line.rfind('=') + 1:].strip()
                    components = re.split(r'\s+', line)
                    for co in components:
                        if co:
                            dependencies.append(co)
    return dependencies


def add_component_remote(base_url, scm, name, origin):
    is_local = add_component_local(name, origin, True)
    if not is_local:
        # clone from the same as name base url
        lib = Repo.fromurl(base_url + '/' + name + '.' + scm.name)
        is_import_success = import_(lib.fullurl, lib.path, True, None, None, False)
        if not is_import_success and base_url != AOS_COMPONENT_BASE_URL:
            # clone from aos component url
            lib = Repo.fromurl(AOS_COMPONENT_BASE_URL + '/' + name + '.' + scm.name)
            is_import_success = import_(lib.fullurl, lib.path, True, None, None, False)

        if not is_import_success:
            error('Component \"%s\" can\'t find in aos && %s && %s' % (name, base_url, AOS_COMPONENT_BASE_URL))

        lib.sync()

        ret_val = change_cube_makefile(lib.name, origin, True, False, True)
        dependencies = get_component_dependencies(os.path.join(lib.path, lib.name + '.mk'))
        for dep in dependencies:
            add_component_remote(lib.url[:lib.url.rfind('/')], lib.scm, dep, get_remote_identify(lib.name))

        out = ('Add remote component \"%s\" success' % lib.url) if ret_val else ('Add remote component \"%s\" fail' % lib.url)
        action(out)
        return ret_val
    else:
        action('Add local component \"%s\" success' % name)
    return True


def get_aos_components(aos_path):
    makefile_dict = {}
    if os.path.isdir(aos_path):
        for (dir_path, dir_names, file_names) in os.walk(aos_path):
            for f in file_names:
                if ('out' not in dir_path) and ('build' not in dir_path) and f.endswith('.mk'):
                    makefile_dict[os.path.join(dir_path, f)] = f[:-3]
    else:
        error('Find components dir is empty!')

    aos_components_dict = {}
    for path, name in makefile_dict.items():
        with open(path, 'r') as f:
            s = f.read()
            component_name = re.findall('\s*NAME\s*:=\s*(\S+)\s*\n', s)
            if len(component_name) == 1:
                aos_components_dict[path.replace(os.path.sep, '/')] = component_name[0]

    return aos_components_dict


def get_remote_identify(name):
    return REMOTE_PATH + '/' + name


def get_local_identify(name):
    dir_name = name.replace('.', os.sep)
    pd = Program(os.getcwd())
    with cd(pd.get_cfg(OS_PATH)):
        if os.path.isfile(os.path.join(dir_name, os.path.basename(dir_name) + '.mk')):
            return dir_name.replace(os.sep, '/')

        aos_paths = os.listdir(os.getcwd())
        for path in aos_paths:
            if os.path.isfile(os.path.join(path, dir_name, os.path.basename(dir_name) + '.mk')):
                return os.path.join(path, dir_name).replace(os.sep, '/')

        aos_components = get_aos_components(pd.get_cfg(OS_PATH))
        for makefile, component in aos_components.items():
            if name == component:
                return os.path.dirname(relpath(pd.get_cfg(OS_PATH), makefile)).replace(os.sep, '/')
    return None


def get_component_identify(name, is_local):
    return get_local_identify(name) if is_local else get_remote_identify(name)


def change_cube_makefile(name, origin, is_add, is_local, is_recursive):
    pd = Program(os.getcwd())
    with cd(pd.get_cfg(PROGRAM_PATH)):
        try:
            with open(CUBE_MAKEFILE, 'r') as fin:
                identify = get_component_identify(name, is_local)
                if not identify:
                    return False

                index = -1
                cube_add_components = 'CUBE_ADD_COMPONENTS :='
                cube_remove_components = 'CUBE_REMOVE_COMPONENTS :='
                lines = []
                cube_modify = False
                for line in fin:
                    line = line.strip()
                    if line.startswith('CUBE_ADD_COMPONENTS :='):
                        cube_add_components = line
                        continue
                    if line.startswith('CUBE_REMOVE_COMPONENTS :='):
                        cube_remove_components = line
                        continue

                    if line.startswith('#'+identify+'='):
                        index = len(lines)
                    lines.append(line)

                if is_add:
                    if identify not in cube_add_components:
                        cube_add_components += ' ' + identify
                        cube_modify = True
                    if identify in cube_remove_components:
                        cube_remove_components = cube_remove_components.replace(identify, '')
                        cube_modify = True

                    if index < 0 and identify != origin:
                        line = '#'+identify+'='+origin
                        lines.append(line)
                        cube_modify = True
                else:
                    if identify not in cube_remove_components:
                        if index >= 0:
                            del lines[index]
                            cube_modify = True
                        else:
                            if not is_recursive:
                                cube_remove_components += ' ' + identify
                                cube_modify = True

                    if identify in cube_add_components:
                        cube_add_components = cube_add_components.replace(identify, '')
                        cube_modify = True
                with open(CUBE_MAKEFILE, 'w') as fout:
                    for line in lines:
                        fout.write(line + '\n')
                    fout.write(cube_add_components + '\n')
                    fout.write(cube_remove_components + '\n')

                if cube_modify:
                    pd.set_cfg(CUBE_MODIFY, '1')
        except (IOError, OSError):
            error('cube.mk isn\'t find current dir')

    return True


def add_component_local(name, origin, is_recursive):
    return change_cube_makefile(name, origin, True, True, is_recursive)


# Add component command
@subcommand('add',
            dict(name='name', help='The component name or URL of the component'),
            dict(name=['-I', '--ignore'], action='store_true', help='Ignore errors related to cloning and updating.'),
            dict(name='--depth', nargs='?',
                 help='Number of revisions to fetch from the remote repository. Default: all revisions.'),
            dict(name='--protocol', nargs='?',
                 help='Transport protocol for the source control management. Supported: https, http, ssh, git. Default: inferred from URL.'),
            help='Add component from AOS_SDK_PATH or URL',
            description=(
                    "Add AliOS-Things component from AOS_SDK_PATH(with dependencies) or GitHub URL(no dependencies) into an existing program.\n"
                    "Use 'aos new TARGET' to new as a program"))
def add(name, path=None, ignore=False, depth=None, protocol=None, top=True):
    cwd_type = Repo().pathtype()
    if Repo.isurl(name):

        lib = Repo.fromurl(name, path)
        if lib.name != OS_NAME and Repo().pathtype() != 'program':
            error('Add component must in AliOS-Things program directory')

        import_(lib.fullurl, lib.path, ignore=ignore, depth=depth, protocol=protocol, top=False)
        lib.sync()

        if lib.name != OS_NAME:
            ret_val = change_cube_makefile(lib.name, get_component_identify(lib.name, False), True, False, False)
            out = ('Add remote component \"%s\" success' % lib.url) if ret_val else ('Add remote component \"%s\" fail' % lib.url)
            action(out)
            dependencies = get_component_dependencies(os.path.join(lib.path, lib.name + '.mk'))
            for dep in dependencies:
                add_component_remote(lib.url[:lib.url.rfind('/')], lib.scm, dep, get_component_identify(lib.name, False))
    else:
        if cwd_type != 'program':
            error('Add component must in AliOS-Things program directory')
        ret_val = add_component_local(name, get_component_identify(name, True), False)
        out = ('Add local component \"%s\" success' % name) if ret_val else ('Add local component \"%s\" fail' % name)
        action(out)


def remove_component_remote(name, origin):
    is_local = get_component_identify(name, True)
    if is_local:
        change_cube_makefile(name, origin, False, True, True)
        out = ('Remove local component \"%s\" success' % name)
    else:
        ret_val = change_cube_makefile(name, origin, False, False, True)
        out = ('Remove remote component \"%s\" success' % name) if ret_val else ('Remove component \"%s\" fail' % name)
        if ret_val:
            pd = Program(os.getcwd())
            dependencies = get_component_dependencies(os.path.join(pd.get_cfg(REMOTE_PATH), name, name + '.mk'))
            for dep in dependencies:
                remove_component_remote(dep, get_component_identify(name, False))

    action(out)


# Remove component
@subcommand('rm',
            dict(name='name', help='Component name'),
            help='Remove component',
            description=(
                    "Remove specified component, its dependencies and references from the current\n"
                    "You can re-add the component from AOS_SDK_PATH or URL via 'aos add name>'."))
def remove(name):
    cwd_type = Repo().pathtype()
    if cwd_type != 'program':
        error('Remove component must in AliOS-Things program dir!')

    ret_val = change_cube_makefile(name, get_component_identify(name, True), False, True, False)
    if ret_val:
        out = ('Remove local component \"%s\" success' % name)
    else:
        ret_val = change_cube_makefile(name, get_component_identify(name, False), False, False, True)
        out = ('Remove remote component \"%s\" success' % name) if ret_val else ('Remove component \"%s\" fail' % name)
        if ret_val:
            pd = Program(os.getcwd())
            dependencies = get_component_dependencies(os.path.join(pd.get_cfg(REMOTE_PATH), name, name + '.mk'))
            for dep in dependencies:
                remove_component_remote(dep, get_component_identify(name, False))

    action(out)


# Deploy command
@subcommand('deploy',
            dict(name=['-I', '--ignore'], action='store_true', help='Ignore errors related to cloning and updating.'),
            dict(name='--depth', nargs='?',
                 help='Number of revisions to fetch from the remote repository. Default: all revisions.'),
            dict(name='--protocol', nargs='?',
                 help='Transport protocol for the source control management. Supported: https, http, ssh, git. Default: inferred from URL.'),
            help='Find and add missing components and source codes',
            description=(
                    "Import missing dependencies in an existing program or component.\n"
                    "Use 'aos import <URL>' and 'aos add <URL>' instead of cloning manually and\n"
                    "then running 'aos deploy'"))
def deploy(ignore=False, depth=None, protocol=None, top=True):
    repo = Repo.fromrepo()
    repo.ignores()

    for lib in repo.libs:
        if os.path.isdir(lib.path):
            if lib.check_repo():
                with cd(lib.path):
                    update(lib.rev, ignore=ignore, depth=depth, protocol=protocol, top=False)
        else:
            import_(lib.fullurl, lib.path, ignore=ignore, depth=depth, protocol=protocol, top=False)
            repo.ignore(relpath(repo.path, lib.path))


# Source command
@subcommand('codes',
            dict(name='name', help='Destination component name without suffix'),
            dict(name=['-I', '--ignore'], action='store_true', help='Ignore errors related to cloning and updating.'),
            dict(name='--depth', nargs='?',
                 help='Number of revisions to fetch from the remote repository. Default: all revisions.'),
            dict(name='--protocol', nargs='?',
                 help='Transport protocol for the source control management. Supported: https, http, ssh, git. Default: inferred from URL.'),
            help='Import the optional component from the remote repository',
            description=(
                    "Import optional component defined by .codes file.\n"
                    "name of the .codes file should be specified ."))
def codes(name, ignore=False, depth=None, protocol=None, top=True):
    repo = Repo.fromrepo()
    code = Repo.fromcode(os.path.join(os.getcwd(), name + '.codes'))

    if os.path.isdir(code.path):
        if code.check_repo():
            with cd(code.path):
                update(code.rev, ignore=ignore, depth=depth, protocol=protocol, top=False)
    else:
        import_(code.fullurl, code.path, ignore=ignore, depth=depth, protocol=protocol, top=False)
        print('>>>>', repo.path, code.path)
        repo.ignore(relpath(repo.path, code.path))


# Publish command
@subcommand('publish',
            dict(name=['-A', '--all'], dest='all_refs', action='store_true',
                 help='Publish all branches, including new ones. Default: push only the current branch.'),
            dict(name=['-M', '--message'], dest='msg', type=str, nargs='?',
                 help='Commit message. Default: prompts for commit message.'),
            help='Publish program or component',
            description=(
                    "Publishes the current program or component and all dependencies to their\nassociated remote repository URLs.\n"
                    "This command performs various consistency checks for local uncommitted changes\n"
                    "and unpublished revisions and encourages to commit/push them.\n"
                    "Online guide about collaboration is available at:\n"
                    "www.aos.com/collab_guide"))
def publish(all_refs=None, msg=None, top=True):
    if top:
        action("Checking for local modifications...")

    repo = Repo.fromrepo()
    if repo.is_local:
        error(
            "%s \"%s\" in \"%s\" with %s is a local repository.\nPlease associate it with a remote repository URL before attempting to publish.\n"
            "Read more about publishing local repositories here:\nhttps://code.aliyun.com/aos/aos-cube/#publishing-local-program-or-component" % (
            "Program" if top else "Library", repo.name, repo.path, repo.scm.name), 1)

    for lib in repo.libs:
        if lib.check_repo():
            with cd(lib.path):
                progress()
                publish(all_refs, msg=msg, top=False)

    sync(recursive=False)

    if repo.dirty():
        action("Uncommitted changes in %s \"%s\" in \"%s\"" % (repo.pathtype(repo.path), repo.name, repo.path))
        if msg:
            repo.commit(msg)
        else:
            raw_input('Press enter to commit and publish: ')
            repo.commit()

    try:
        outgoing = repo.outgoing()
        if outgoing > 0:
            action("Pushing local repository \"%s\" to remote \"%s\"" % (repo.name, repo.url))
            repo.publish(all_refs)
        else:
            if top:
                action("Nothing to publish to the remote repository (the source tree is unmodified)")
    except ProcessException as e:
        if e[0] != 1:
            raise e


# Update command
@subcommand('update',
            dict(name='rev', nargs='?', help='Revision, tag or branch'),
            dict(name=['-C', '--clean'], action='store_true',
                 help='Perform a clean update and discard all modified or untracked files. WARNING: This action cannot be undone. Use with caution.'),
            dict(name='--clean-files', action='store_true',
                 help='Remove any local ignored files. Requires \'--clean\'. WARNING: This will wipe all local uncommitted, untracked and ignored files. Use with extreme caution.'),
            dict(name='--clean-deps', action='store_true',
                 help='Remove any local libraries and also libraries containing uncommitted or unpublished changes. Requires \'--clean\'. WARNING: This action cannot be undone. Use with caution.'),
            dict(name=['-I', '--ignore'], action='store_true',
                 help='Ignore errors related to unpublished libraries, unpublished or uncommitted changes, and attempt to update from associated remote repository URLs.'),
            dict(name='--depth', nargs='?',
                 help='Number of revisions to fetch from the remote repository. Default: all revisions.'),
            dict(name='--protocol', nargs='?',
                 help='Transport protocol for the source control management. Supported: https, http, ssh, git. Default: inferred from URL.'),
            help='Update to branch, tag, revision or latest',
            description=(
                    "Updates the current program or component and its dependencies to specified\nbranch, tag or revision.\n"
                    "Alternatively fetches from associated remote repository URL and updates to the\n"
                    "latest revision in the current branch."))
def update(rev=None, clean=False, clean_files=False, clean_deps=False, ignore=False, top=True, depth=None,
           protocol=None):
    if top and clean:
        sync()

    cwd_type = Repo.pathtype(cwd_root)
    cwd_dest = "program" if cwd_type == "directory" else "component"

    repo = Repo.fromrepo()
    # A copy of repo containing the .lib layout before updating
    repo_orig = Repo.fromrepo()

    if top and not rev and repo.isdetached():
        error(
            "This %s is in detached HEAD state, and you won't be able to receive updates from the remote repository until you either checkout a branch or create a new one.\n"
            "You can checkout a branch using \"%s checkout <branch_name>\" command before running \"aos update\"." % (
            cwd_type, repo.scm.name), 1)

    if repo.is_local and not repo.rev:
        action("Skipping unpublished empty %s \"%s\"" % (
            cwd_type if top else cwd_dest,
            os.path.basename(repo.path) if top else relpath(cwd_root, repo.path)))
    else:
        # Fetch from remote repo
        action("Updating %s \"%s\" to %s" % (
            cwd_type if top else cwd_dest,
            os.path.basename(repo.path) if top else relpath(cwd_root, repo.path),
            repo.revtype(rev, True)))

        try:
            repo.update(rev, clean, clean_files, repo.is_local)
        except ProcessException as e:
            err = "Unable to update \"%s\" to %s" % (repo.name, repo.revtype(rev, True))
            if depth:
                err = err + (
                        "\nThe --depth option might prevent fetching the whole revision tree and checking out %s." % (
                repo.revtype(repo.rev, True)))
            if ignore:
                warning(err)
            else:
                error(err, e[0])

        repo.rm_untracked()
        if top and cwd_type == 'component':
            repo.sync()
            repo.write()

    # Compare component references (.lib) before and after update, and remove libraries that do not have references in the current revision
    for lib in repo_orig.libs:
        if not os.path.isfile(lib.lib) and os.path.isdir(
                lib.path):  # Library reference doesn't exist in the new revision. Will try to remove component to reproduce original structure
            gc = False
            with cd(lib.path):
                lib_repo = Repo.fromrepo(lib.path)
                gc, msg = lib_repo.can_update(clean, clean_deps)
            if gc:
                action("Remove component \"%s\" (obsolete)" % (relpath(cwd_root, lib.path)))
                rmtree_readonly(lib.path)
                repo.unignore(relpath(repo.path, lib.path))
            else:
                if ignore:
                    warning(msg)
                else:
                    error(msg, 1)

    # Reinitialize repo.libs() to reflect the component files after update
    repo.sync()

    # Recheck libraries as their urls might have changed
    for lib in repo.libs:
        if os.path.isdir(lib.path) and Repo.isrepo(lib.path):
            lib_repo = Repo.fromrepo(lib.path)
            if (not lib.is_local and not lib_repo.is_local and
                    formaturl(lib.url, 'https') != formaturl(lib_repo.url, 'https')):  # Repository URL has changed
                gc = False
                with cd(lib.path):
                    gc, msg = lib_repo.can_update(clean, clean_deps)
                if gc:
                    action(
                        "Remove component \"%s\" (changed URL). Will add from new URL." % (relpath(cwd_root, lib.path)))
                    rmtree_readonly(lib.path)
                    repo.unignore(relpath(repo.path, lib.path))
                else:
                    if ignore:
                        warning(msg)
                    else:
                        error(msg, 1)

    # Import missing repos and update to revs
    for lib in repo.libs:
        if not os.path.isdir(lib.path):
            import_(lib.fullurl, lib.path, ignore=ignore, depth=depth, protocol=protocol, top=False)
            repo.ignore(relpath(repo.path, lib.path))
        else:
            with cd(lib.path):
                update(lib.rev, clean=clean, clean_files=clean_files, clean_deps=clean_deps, ignore=ignore, top=False)


#    if top:
#        program = Program(repo.path)
#        program.set_root()
#        program.post_action()
#        if program.is_classic:
#            program.update_tools('.temp')


# Synch command
@subcommand('sync',
            help='Synchronize aos component references',
            description=(
                    "Synchronizes all component and dependency references (.component files) in the\n"
                    "current program or component.\n"
                    "Note that this will remove all invalid component references."))
def sync(recursive=True, keep_refs=False, top=True):
    if top and recursive:
        action("Synchronizing dependency references...")

    repo = Repo.fromrepo()
    repo.ignores()

    for lib in repo.libs:
        if os.path.isdir(lib.path):
            lib.check_repo()
            lib.sync()
            lib.write()
            repo.ignore(relpath(repo.path, lib.path))
            progress()
        else:
            if not keep_refs:
                action("Remove reference \"%s\" -> \"%s\"" % (lib.name, lib.fullurl))
                repo.remove(lib.lib)
                repo.unignore(relpath(repo.path, lib.path))

    for code in repo.codes:
        if os.path.isdir(code.path):
            code.check_repo()
            code.sync()
            code.write_codes()
            repo.ignore(relpath(repo.path, code.path))
            progress()

    for root, dirs, files in os.walk(repo.path):
        dirs[:] = [d for d in dirs if not d.startswith('.')]
        files[:] = [f for f in files if not f.startswith('.')]

        for d in list(dirs):
            if not Repo.isrepo(os.path.join(root, d)):
                continue

            lib = Repo.fromrepo(os.path.join(root, d))
            if os.path.isfile(lib.lib) or os.path.isfile(lib.code):
                dirs.remove(d)
                continue

            dirs.remove(d)
            lib.write()
            repo.ignore(relpath(repo.path, lib.path))
            repo.add(lib.lib)
            progress()

    repo.sync()

    if recursive:
        for lib in repo.libs:
            if lib.check_repo():
                with cd(lib.path):
                    sync(keep_refs=keep_refs, top=False)

    # Update the .lib reference in the parent repository
    cwd_type = Repo.pathtype(cwd_root)
    if top and cwd_type == "component":
        repo = Repo.fromrepo()
        repo.write()


# Command status for cross-SCM status of repositories
@subcommand('status',
            dict(name=['-I', '--ignore'], action='store_true', help='Ignore errors related to missing libraries.'),
            help='Show version control status\n\n',
            description=(
                    "Show uncommitted changes a program or component and its dependencies."))
def status_(ignore=False):
    repo = Repo.fromrepo()
    if repo.dirty():
        action("Status for \"%s\":" % repo.name)
        log(repo.status() + "\n")

    for lib in repo.libs:
        if lib.check_repo(ignore):
            with cd(lib.path):
                status_(ignore)

def make_build(make_args):
    # aos new program
    if os.path.isfile(os.path.join(os.getcwd(), Cfg.file)):
        pd = Program(os.getcwd())
        os_path = pd.get_cfg(OS_PATH)

        build_dir = ''
        out_path = ''
        is_build_dir = 'BUILD_DIR=' in make_args or os.environ.get('BUILD_DIR')
        if not is_build_dir:
            out_path = os.path.join(os.getcwd(), 'out').replace(os.path.sep, '/')
            build_dir = 'BUILD_DIR=' + out_path

        # cube_makefile = 'CUBE_MAKEFILE=' + os.path.join(pd.get_cfg(PROGRAM_PATH), CUBE_MAKEFILE).replace(os.path.sep, '/')
        app_dir = 'APPDIR=' + os.path.dirname(os.getcwd())

        with cd(os_path):
            cube_modify = pd.get_cfg(CUBE_MODIFY)
            if cube_modify == '1':
                _run_make(['-e', '-f build/Makefile', 'clean'])
                if build_dir:
                    shutil.rmtree(out_path, True)
                pd.set_cfg(CUBE_MODIFY, '0')
            _run_make(['-e', '-f build/Makefile', make_args, app_dir, build_dir])
    else:
        # aos source code
        _run_make(['-e', '-f build/Makefile', make_args])

def scons_build(args):
    make_args = ['scons -j4 -f ucube.py']
    target_find = False

    for arg in args:
        if '@' in arg and not target_find:
            targets = arg.split('@')
            if len(targets) < 2:
                error('Must special app and board when build aos')

            app = 'APPLICATION='+targets[0]
            board = 'BOARD=' + targets[1]
            make_args.append(app)
            make_args.append(board)

            if len(targets) == 3:
                build_type = 'TYPE=' + targets[2]
                make_args.append(build_type)

            target_find = True
        else:
            make_args.append(arg)

    popen(' '.join(make_args), shell=True, cwd=os.getcwd())


def get_scons_enabled_boards():
    if os.path.exists("build/scons_enabled.py"):
        module_path = os.path.abspath(os.path.join('build'))
        sys.path.append(module_path)
        from scons_enabled import scons_enabled_boards
        return scons_enabled_boards
    else:
        return []

# Make command
@subcommand('make',
            help='Make aos program/component',
            description="Make aos program/component.")
def make():
    from .constant import ver
    log('aos-cube version: %s\n' %ver)

    # Get the make arguments
    args = sys.argv[2:]


    make_args = ' '.join(args)
    for arg in args:
        if '@' in arg:
            targets = arg.split('@')
            if len(targets) < 2:
                error('Must special app and board when build aos')

            board = targets[1]
            if board in get_scons_enabled_boards():
                scons_build(args)
            else:
                make_build(make_args)

            return
    #aos make clean go here
    make_build(make_args)


# scons make command
@subcommand('scons',
            help='Make aos program/component by scons',
            description="Make aos program/component by scons.")
def scons():
    # Get the make arguments
    args = sys.argv[2:]

    #install dependent toolchains
    _install_toolchains(args)

    scons_build(args)


# Make command
@subcommand('makelib',
            dict(name='path', nargs='?', help='Library directory path.'),
            dict(name=['-N', '--new'], dest='new', action='store_true', help='Create a library configure makefile .'),
            dict(name=['-r', '--arch'],
                 help='The arch for static library compiling. E.g. "Cortex-M3", "Cortex-M4", "Cortex-M4F", "ARM968E-S", "linux"'),
            help='Compile static library\n\n',
            description=(
                    "Compile static library."))
def makelib(path, arch=None, new=False):
    if new:
        root_dir = Program().path
        with open(os.path.join(root_dir, 'build/template.mk'), 'r') as f:
            content = f.read()
        with open(os.path.join(root_dir, path, os.path.basename(path) + '_src.mk'), 'w') as f:
            f.write(content.replace('template', os.path.basename(path)))
        return

    if arch:
        _run_make(['-e', 'LIB_DIR=' + path, 'TARGET_ARCH=' + arch, '-f build/aos_library_makefile.mk'])
        return

    host_arch = ['Cortex-M3', 'Cortex-M4', 'Cortex-M4F', 'ARM968E-S', 'linux']
    for arch in host_arch:
        _run_make(['-e', 'LIB_DIR=' + path, 'TARGET_ARCH=' + arch, '-f build/aos_library_makefile.mk'])


# Generic config command
@subcommand('config',
            dict(name='var', nargs='?', help='Variable name. E.g. "AOS_SDK_PATH"'),
            dict(name='value', nargs='?',
                 help='Value. Will show the currently set default value for a variable if not specified.'),
            dict(name=['-g', '--global'], dest='global_cfg', action='store_true',
                 help='Use global settings, not local'),
            dict(name=['-u', '--unset'], dest='unset', action='store_true', help='Unset the specified variable.'),
            dict(name=['-l', '--list'], dest='list_config', action='store_true', help='List aos tool configuration.'),
            help='Tool configuration\n\n',
            description=(
                    "Gets, sets or unsets aos tool configuration options.\n"
                    "Options can be global (via the --global switch) or local (per program)\n"
                    "Global options are always overridden by local/program options.\n"
                    "Currently supported options: aosder, protocol, depth, cache"))

def config_(var=None, value=None, global_cfg=False, unset=False, list_config=False):
    name = var
    var = str(var).upper()

    if list_config:
        g = Global()
        g_vars = g.list_cfg().items()
        action("Global config:")
        if g_vars:
            for v in g_vars:
                log("%s=%s\n" % (v[0], v[1]))
        else:
            log("No global configuration is set\n")
        log("\n")

        p = Program(os.getcwd())
        action("Local config (%s):" % p.path)
        if not p.is_cwd:
            p_vars = p.list_cfg().items()
            if p_vars:
                for v in p_vars:
                    log("%s=%s\n" % (v[0], v[1]))
            else:
                log("No local configuration is set\n")
        else:
            log("Couldn't find valid aos program in %s\n" % p.path)

    elif name:
        if global_cfg:
            # Global configuration
            g = Global()
            if unset:
                g.set_cfg(var, None)
                action('Unset global %s' % name)
            elif value:
                g.set_cfg(var, value)
                action('%s now set as global %s' % (value, name))
            else:
                value = g.get_cfg(var)
                action(('%s' % value) if value else 'No global %s set' % (name))
        else:
            # Find the root of the program
            program = Program(os.getcwd())
            if program.is_cwd and not var == 'ROOT':
                error(
                    "Could not find aos program in current path \"%s\".\n"
                    "Change the current directory to a valid aos program, set the current directory as an aos program with 'aos config root .', or use the '--global' option to set global configuration." % program.path)
            with cd(program.path):
                if unset:
                    program.set_cfg(var, None)
                    action('Unset default %s in program "%s"' % (name, program.name))
                elif value:
                    program.set_cfg(var, value)
                    action('%s now set as default %s in program "%s"' % (value, name, program.name))
                else:
                    value = program.get_cfg(var)
                    action(('%s' % value) if value else 'No default %s set in program "%s"' % (name, program.name))
    else:
        subcommands['config'].error("too few arguments")

def scons_upload(args):
    make_args = ['scons -f ucube.py COMMAND=upload']
    target_find = False

    for arg in args:
        if '@' in arg and not target_find:
            targets = arg.split('@')
            if len(targets) < 2:
                error('Must special app and board when build aos')

            app = 'APPLICATION='+targets[0]
            board = 'BOARD=' + targets[1]
            make_args.append(app)
            make_args.append(board)

            if len(targets) == 3:
                build_type = 'TYPE=' + targets[2]
                make_args.append(build_type)

            target_find = True
        #else:
        #    ret = make_args.append(arg)

    ret = popen(' '.join(make_args), shell=True, cwd=os.getcwd())
#    ret = subprocess.call(make_args, stdout=sys.stdout, stderr=sys.stderr)
    return ret

def _run_upload(target, aos_path):
    from .constant import ver
    log('aos-cube version: %s\n' %ver)
    # Get the make arguments
    args = sys.argv[2:]
    ret = scons_upload(args)
    return ret



# upload command
@subcommand('upload',
            dict(name='target', help='aos target, e.g. helloworld@starterkit'),
            help='Upload aos image',
            description="Upload aos image to target platform.")

def upload(target):
    from .constant import ver
    log('aos-cube version: %s\n' %ver)

    log("[INFO]:target: %s\n" % target)

    if os.path.isdir('./include/aos') == False:
        log("[INFO]:Not in aos_sdk_path, curr_path:'%s'\n" % os.getcwd())
        aos_path = Global().get_cfg(AOS_SDK_PATH)
        if aos_path == None:
            error("[ERRO]:Not in aos_sdk_path, aos_sdk unavailable as well. Please run 'aos new $prj_name'!")
        else:
            log("[INFO]:Config Loading OK, use '%s' as sdk path\n" % aos_path)
    else:
        aos_path = os.getcwd()
        log("[INFO]:Currently in aos_sdk_path: '%s'\n" % os.getcwd())

    if '@' not in target or len(target.split('@')) != 2:
        error('target invalid')
        return;

    _run_upload(target, aos_path)



@subcommand('monitor',
            dict(name='port', help='Choose serial port'),
            dict(name='baud', help='Select baudrate'),
            help='Serial port monitor\n\n',
            description="Open a simple serial monitor.")
def monitor(port, baud):
    args = ['miniterm', port, baud]
    host_os = get_host_os()
    if host_os != 'Win32':
        args.append('--eol')
        args.append('LF')
    sys.argv = args
    try:
        miniterm.main()
    except ProcessException as e:
        raise e

@subcommand('devices',
            help='List devices on serial ports\n\n',
            description="List devices on serial ports.")
def devices():
    arr = []
    for p in comports():
        j = json.dumps(p.__dict__)
        arr.append(json.loads(j))
    print(json.dumps(arr, indent = 4))
    sys.exit(0)

@subcommand('upgrade',
            help='Upgrade aos-cube to latest',
            description="Run pip upgrade process to keep aos-cube up-to-date.")
def upgrade():
    cmd = ["pip", "install", "--upgrade", "aos-cube"]
    try:
        ret = popen(cmd)
        if ret != 0:
            cmd.insert(3, "--no-cache-dir")
            popen(cmd)
    except Exception as e:
        raise e

@subcommand('help',
            help='This help screen')
def help_():
    return parser.print_help()


def main():
    global verbose, very_verbose, remainder, cwd_root

    reload(sys)
    sys.setdefaultencoding('UTF8')

    # Help messages adapt based on current dir
    cwd_root = os.getcwd()

    if sys.version_info[0] != 2 or sys.version_info[1] < 7:
        error(
            "aos cube is compatible with Python version >= 2.7 and < 3.0\n"
            "Please refer to the online guide available at https://code.aliyun.com/aos/aos-cube")

    # Parse/run command
    if len(sys.argv) <= 1:
        help_()
        sys.exit(1)

    if '--version' in sys.argv:
        log(ver + "\n")
        sys.exit(0)

    pargs, remainder = parser.parse_known_args()
    status = 1

    try:
        very_verbose = pargs.very_verbose
        verbose = very_verbose or pargs.verbose
        info('Working path \"%s\" (%s)' % (os.getcwd(), Repo.pathtype(cwd_root)))
        status = pargs.command(pargs)
    except ProcessException as e:
        error(
            "\"%s\" returned error code %d.\n"
            "Command \"%s\" in \"%s\"" % (e[1], e[0], e[2], e[3]), e[0])
    except OSError as e:
        traceback.print_exc()
        if e[0] == errno.ENOENT:
            error(
                "Could not detect one of the command-line tools.\n"
                "You could retry the last command with \"-v\" flag for verbose output\n", e[0])
        else:
            error('OS Error: %s' % e[1], e[0])
    except KeyboardInterrupt as e:
        info('User aborted!', -1)
        sys.exit(255)
    except Exception as e:
        traceback.print_exc(file=sys.stdout)
        if very_verbose:
            traceback.print_exc(file=sys.stdout)
        error("Unknown Error: %s" % e, 255)
    sys.exit(status or 0)


if __name__ == "__main__":
    main()



import os, sys, time, serial, subprocess, traceback, glob

def list_devices(host_os):
    return glob.glob('/dev/esp32-*')

def exist(device):
    return os.path.exists(device)

def new_device(device):
    try:
        ser = serial.Serial(device, 115200, timeout = 0.02)
        ser.setRTS(True)
        ser.setDTR(False)
        time.sleep(0.1)
        ser.setDTR(True)
    except:
        ser = None
        print 'esp32: open {0} error'.format(device)
    return ser

def erase(device):
    retry = 3
    baudrate = 230400
    error = 'fail'
    while retry > 0:
        script = ['esptool.py']
        script += ['--chip']
        script += ['esp32']
        script += ['--port']
        script += [device]
        script += ['--baud']
        script += [str(baudrate)]
        script += ['erase_flash']
        ret = subprocess.call(script)
        if ret == 0:
            error = 'success'
            break
        retry -= 1
        baudrate = baudrate / 2
    return error

def program(device, address, file):
    retry = 3
    baudrate = 230400
    error = 'fail'
    flash_tool_path = os.path.dirname(os.path.realpath(__file__)) + '/esptool.py'
    while retry > 0:
        script = ['esptool.py']
        script += ['--chip']
        script += ['esp32']
        script += ['--port']
        script += [device]
        script += ['--baud']
        script += [str(baudrate)]
        script += ['--before']
        script += ['default_reset']
        script += ['write_flash']
        script += ['-z']
        script += ['--flash_mode']
        script += ['dio']
        script += ['--flash_freq']
        script += ['40m']
        script += ['--flash_size']
        script += ['4MB']
        script += [address]
        script += [file]
        ret = subprocess.call(script)
        if ret == 0:
            error =  'success'
            break
        retry -= 1
        baudrate = baudrate / 2
    control(device, 'reset')
    return error

def control(device, operation):
    try:
        ser = serial.Serial(device, 115200)
    except:
        traceback.print_exc()
        print 'esp32 control error: unable to open {0}'.format(device)
        return 'fail'
    ret = 'fail'
    try:
        if operation == 'reset':
            ser.setDTR(False)
            time.sleep(0.1)
            ser.setDTR(True)
            ret = 'success'
        elif operation == 'stop':
            ser.setDTR(False)
            ret = 'success'
        elif operation == 'start':
            ser.setDTR(True)
            ret = 'success'
    except:
        pass
    ser.close()
    return ret
